# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

# Imports
import os

# DendroTweaks imports
from dendrotweaks.stimuli.populations import Population
from dendrotweaks.stimuli.iclamps import IClamp
from dendrotweaks.morphology.io import create_segment_tree
from dendrotweaks.prerun import prerun
from dendrotweaks.utils import calculate_lambda_f

# Warnings configuration
import warnings

def custom_warning_formatter(message, category, filename, lineno, file=None, line=None):
    return f"WARNING: {message}\n({os.path.basename(filename)}, line {lineno})\n"

warnings.formatwarning = custom_warning_formatter


class SimulationMixin:

    """Mixin class for model simulation functionalities."""

    @property
    def recordings(self):
        """
        The recordings of the model. Reference to the recordings in the simulator.
        """
        return self.simulator.recordings


    @recordings.setter
    def recordings(self, recordings):
        self.simulator.recordings = recordings


    # ========================================================================
    # SEGMENTATION
    # ========================================================================

    # TODO Make a context manager for this
    def _temp_clear_stimuli(self):
        """
        Temporarily save and clear stimuli.
        """
        self.export_stimuli(file_name='_temp_stimuli')
        self.remove_all_stimuli()
        self.remove_all_recordings()


    def _temp_reload_stimuli(self):
        """
        Load stimuli from a temporary file and clean up.
        """
        self.load_stimuli(file_name='_temp_stimuli')
        self.path_manager.remove_folder('stimuli/_temp_stimuli')


    def set_segmentation(self, d_lambda=0.1, f=100):
        """
        Set the number of segments in each section based on the geometry.

        Parameters
        ----------
        d_lambda : float
            The lambda value to use.
        f : float
            The frequency value to use.
        """
        self.d_lambda = d_lambda

        # Temporarily save and clear stimuli
        self._temp_clear_stimuli()

        # Pre-distribute parameters needed for lambda_f calculation
        for param_name in ['cm', 'Ra']:
            self.distribute(param_name)

        # Calculate lambda_f and set nseg for each section
        for sec in self.sec_tree.sections:
            lambda_f = calculate_lambda_f(sec.distances, sec.diameters, sec.Ra, sec.cm, f)
            nseg = max(1, int((sec.L / (d_lambda * lambda_f) + 0.9) / 2) * 2 + 1)
            sec._nseg = sec._ref.nseg = nseg

        # Rebuild the segment tree and redistribute parameters
        self.seg_tree = create_segment_tree(self.sec_tree)
        self.distribute_all()

        # Reload stimuli and clean up temporary files
        self._temp_reload_stimuli()


    # -----------------------------------------------------------------------
    # ICLAMPS
    # -----------------------------------------------------------------------

    def add_iclamp(self, sec, loc, amp=0, delay=100, dur=100):
        """
        Add an IClamp to a section.

        Parameters
        ----------
        sec : Section
            The section to add the IClamp to.
        loc : float
            The location of the IClamp in the section.
        amp : float, optional
            The amplitude of the IClamp. Default is 0.
        delay : float, optional
            The delay of the IClamp. Default is 100.
        dur : float, optional
            The duration of the IClamp. Default is 100.
        """
        seg = sec(loc)
        if self.iclamps.get(seg):
            self.remove_iclamp(sec, loc)
        iclamp = IClamp(sec, loc, amp, delay, dur)
        if self.verbose: print(f'IClamp added to sec {sec} at loc {loc}.')
        self.iclamps[seg] = iclamp


    def remove_iclamp(self, sec, loc):
        """
        Remove an IClamp from a section.

        Parameters
        ----------
        sec : Section
            The section to remove the IClamp from.
        loc : float
            The location of the IClamp in the section.
        """
        seg = sec(loc)
        if self.iclamps.get(seg):
            self.iclamps.pop(seg)


    def remove_all_iclamps(self):
        """
        Remove all IClamps from the model.
        """

        for seg in list(self.iclamps.keys()):
            sec, loc = seg._section, seg.x
            self.remove_iclamp(sec, loc)
        if self.iclamps:
            warnings.warn(f'Not all iclamps were removed: {self.iclamps}')
        self.iclamps = {}


    # -----------------------------------------------------------------------
    # SYNAPSES
    # -----------------------------------------------------------------------

    def _add_population(self, population):
        self.populations[population.name] = population


    def add_population(self, name, segments, N, syn_type):
        """
        Add a population of synapses to the model.

        Parameters
        ----------
        name : str
            The name of the population.
        segments : list[Segment]
            The segments to add the synapses to.
        N : int
            The number of synapses to add.
        syn_type : str
            The type of synapse to add.
        """
        population = Population(name, segments, N, syn_type)
        population.allocate_synapses()
        population.create_inputs()
        self._add_population(population)


    def update_population_kinetic_params(self, pop_name, **params):
        """
        Update the kinetic parameters of a population of synapses.

        Parameters
        ----------
        pop_name : str
            The name of the population.
        params : dict
            The parameters to update.
        """
        population = self.populations[pop_name]
        population.update_kinetic_params(**params)
        print(population.kinetic_params)

    
    def update_population_input_params(self, pop_name, **params):
        """
        Update the input parameters of a population of synapses.

        Parameters
        ----------
        pop_name : str
            The name of the population.
        params : dict
            The parameters to update.
        """
        population = self.populations[pop_name]
        population.update_input_params(**params)
        print(population.input_params)


    def remove_population(self, name):
        """
        Remove a population of synapses from the model.

        Parameters  
        ----------
        name : str
            The name of the population
        """
        population = self.populations.pop(name)
        population.clean()
        

    def remove_all_populations(self):
        """
        Remove all populations of synapses from the model.
        """
        for name in list(self.populations.keys()):
            self.remove_population(name)
        if any(self.populations.values()):
            warnings.warn(f'Not all populations were removed: {self.populations}')
        self.populations = {}


    def remove_all_stimuli(self):
        """
        Remove all stimuli from the model.
        """
        self.remove_all_iclamps()
        self.remove_all_populations()


    # ========================================================================
    # SIMULATION
    # ========================================================================

    def add_recording(self, sec, loc, var='v'):
        """
        Add a recording to the model.

        Parameters
        ----------
        sec : Section
            The section to record from.
        loc : float
            The location along the normalized section length to record from.
        var : str, optional
            The variable to record. Default is 'v'.
        """
        self.simulator.add_recording(sec, loc, var)
        if self.verbose: print(f'Recording added to sec {sec} at loc {loc}.')


    def remove_recording(self, sec, loc, var='v'):
        """
        Remove a recording from the model.

        Parameters
        ----------
        sec : Section
            The section to remove the recording from.
        loc : float
            The location along the normalized section length to remove the recording from.
        """
        self.simulator.remove_recording(sec, loc, var)


    def remove_all_recordings(self, var=None):
        """
        Remove all recordings from the model.
        """
        self.simulator.remove_all_recordings(var=var)


    def run(self, duration=300, prerun_time=0, truncate=True):
        """
        Run the simulation for a specified duration, optionally preceded by a prerun period
        to stabilize the model.

        Parameters
        ----------
        duration : float
            Duration of the main simulation (excluding prerun).
        prerun_time : float
            Optional prerun period to run before the main simulation.
        truncate : bool
            Whether to truncate prerun data after the simulation.
        """
        if duration <= 0:
            raise ValueError("Simulation duration must be positive.")
        if prerun_time < 0:
            raise ValueError("Prerun time must be non-negative.")

        total_time = duration + prerun_time

        if prerun_time > 0:
            with prerun(self, duration=prerun_time, truncate=truncate):
                self.simulator.run(total_time)
        else:
            self.simulator.run(duration)


    def get_traces(self):
        return self.simulator.get_traces()


    def plot(self, *args, **kwargs):
        self.simulator.plot(*args, **kwargs)


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from scipy.signal import find_peaks, peak_widths
from scipy.optimize import curve_fit


# =============================================================================
# PASSIVE PROPERTIES
# =============================================================================

def get_somatic_data(model):
    """
    Get the somatic voltage, time, time step, and injected current.

    Parameters
    ----------
    model : Model
        The neuron model.

    Returns
    -------
    tuple
        A tuple containing the voltage, time, time step, and injected current.
    """
    soma = model.sec_tree.root
    seg = soma(0.5)
    iclamp = model.iclamps[seg]

    v = np.array(model.simulator.recordings['v'][seg])
    t = np.array(model.simulator.t)
    dt = model.simulator.dt

    return v, t, dt, iclamp


def calculate_input_resistance(model):
    """
    Calculate the input resistance of the neuron model.
        
    This function determines the input resistance by calculating the ratio 
    between the voltage change and the injected current. The voltage change 
    is measured as the difference between the membrane potential at the onset 
    and offset of the current injection.

    Parameters
    ----------
    model : Model
        The neuron model.

    Returns
    -------
    dict
        A dictionary containing the onset and offset voltages, the input resistance, and the injected current.
    """
    
    v, t, dt, iclamp = get_somatic_data(model)

    amp = iclamp.amp
    start_ts = int(iclamp.delay / dt)
    stop_ts = int((iclamp.delay + iclamp.dur) / dt)
    v_onset = v[int(start_ts)]
    v_offset = v[int(stop_ts)]
    
    R_in = (v_offset - v_onset) / amp
    print(f"Input resistance: {R_in:.2f} MOhm")

    return {
        'onset_voltage': v_onset,
        'offset_voltage': v_offset,
        'input_resistance': R_in,
        'current_amplitude': amp
    }


def _double_exp_decay(t, A1, tau1, A2, tau2):
    return A1 * np.exp(-t / tau1) + A2 * np.exp(-t / tau2)

def calculate_time_constant(model):
    """
    Estimate the passive membrane time constant (τm) by fitting
    a double exponential to the somatic voltage decay and selecting
    the slowest τ component.

    Parameters
    ----------
    model : Model
        The neuron model (assumes Ih is disabled).

    Returns
    -------
    dict
        A dictionary with τm (slowest), both τs, and fit details.
    """
    v, t, dt, iclamp = get_somatic_data(model)

    start_ts = int(iclamp.delay / dt)
    stop_ts = int((iclamp.delay + iclamp.dur) / dt)
    min_ts = np.argmin(v[start_ts:stop_ts]) + start_ts
    v_min = v[min_ts]
    v_decay = v[start_ts: min_ts] - v_min
    t_decay = t[start_ts: min_ts] - t[start_ts]

    # Fit double exponential
    try:
        popt, _ = curve_fit(
            _double_exp_decay, t_decay, v_decay, 
            p0=[1, 10, 0.5, 100], 
            bounds=(0, [np.inf, 1000, np.inf, 1000])
        )
        A1, tau1, A2, tau2 = popt
        tau_slowest = max(tau1, tau2)
    except RuntimeError:
        print("Fit failed. Could not estimate time constant.")
        return None

    print(f"Time constant {tau_slowest:.2f} ms. Estimated from double exp fit (slowest component)")

    return {
        'time_constant': tau_slowest,
        'start_time': start_ts * dt,
        'tau1': tau1,
        'tau2': tau2,
        'A1': A1,
        'A2': A2,
        'decay_time': t_decay,
        'decay_voltage': v_decay
    }


def calculate_passive_properties(model):
    """
    Calculate the passive properties of the neuron model.

    Parameters
    ----------
    model : Model
        The neuron model.   

    Returns
    -------
    dict
        A dictionary containing the input resistance, time constant, and the exponential fit.
    """

    data_rin = calculate_input_resistance(model)
    data_tau = calculate_time_constant(model)

    return {**data_rin, **data_tau}

def plot_passive_properties(data, ax=None):
        
    if ax is None:
        _, ax = plt.subplots()

    R_in = data['input_resistance']
    tau = data['time_constant']
    v_onset = data['onset_voltage']
    v_offset = data['offset_voltage']
    t_decay = data['decay_time']
    v_decay = data['decay_voltage']
    start_t = data['start_time']

    ax.set_title(f"R_in: {R_in:.2f} MOhm, Tau: {tau:.2f} ms")
    ax.axhline(v_onset, color='gray', linestyle='--', label='V onset')
    ax.axhline(v_offset, color='gray', linestyle='--', label='V offset')
    
    # Shift the exp_decay output along the y-axis
    fit_curve = _double_exp_decay(t_decay, data['A1'], data['tau1'], data['A2'], data['tau2']) + v_offset
    label = f'Double exp fit (tau1 = {data["tau1"]:.1f} ms, tau2 = {data["tau2"]:.1f} ms)'

    ax.plot(t_decay + start_t, fit_curve, color='red', label='Exp. fit', linestyle='--')
    ax.legend()


# =============================================================================
# ACTIVE PROPERTIES
# =============================================================================

def detect_somatic_spikes(model, **kwargs):
    """
    Detect somatic spikes in the model and calculate spike amplitudes and widths.
    
    Returns:
        dict: A dictionary containing spike metrics.
    """
    soma = model.sec_tree.root
    seg = soma(0.5)
            
    v = np.array(model.simulator.recordings['v'][seg])
    t = np.array(model.simulator.t)
    dt = model.simulator.dt

    baseline = np.median(v)
    height = kwargs.get('height', baseline)
    distance = kwargs.get('distance', int(2/dt))
    prominence = kwargs.get('prominence', 50)
    wlen = kwargs.get('wlen', int(20/dt))
    
    peaks, properties = find_peaks(v, height=height, distance=distance, prominence=prominence, wlen=wlen)
    half_widths, _, left_bases, right_bases = peak_widths(v, peaks, rel_height=0.5)
    half_widths *= dt
    left_bases *= dt
    right_bases *= dt

    return {
        'spike_times': t[peaks],
        'spike_values': properties['peak_heights'],
        'half_widths': half_widths,
        'amplitudes': properties['prominences'],
        'left_bases': left_bases,
        'right_bases': right_bases,
        'stimulus_duration': model.iclamps[seg].dur
    }


def plot_somatic_spikes(data, ax=None, show_metrics=False):
    """Plot detected spikes on the provided axis or create a new figure.
    
    Args:
        model: The neuron model
        ax: Optional matplotlib axis for plotting
        
    Returns:
        matplotlib.axes.Axes: The plot axis
    """

    spike_times = data['spike_times']
    spike_values = data['spike_values']
    half_widths = data['half_widths']
    amplitudes = data['amplitudes']
    right_bases = data['right_bases']
    left_bases = data['left_bases']
    duration_ms = data['stimulus_duration']

    n_spikes = len(spike_times)

    if n_spikes == 0:
        return    

    print(f"Detected {n_spikes} spikes")
    print(f"Average spike half-width: {np.mean(half_widths):.2f} ms")
    print(f"Average spike amplitude: {np.mean(amplitudes):.2f} mV")
    print(f"Spike frequency: {n_spikes / duration_ms * 1000:.2f} Hz")
    
    ax.plot(spike_times, spike_values, 'o', color='red')
    ax.set_xlabel('Time (ms)')
    ax.set_ylabel('Amplitude (mV)')
    ax.set_title(f'Somatic spikes ({len(spike_times)} detected)')
    
    if show_metrics:
        for t, v, w, a, lb, rb in zip(spike_times, spike_values, half_widths, amplitudes, left_bases, right_bases):
            # plot spike amplitude
            ax.plot([t, t], [v, v - a], color='red', linestyle='--')
            # plot spike width
            ax.plot([t - 10*w/2, t + 10*w/2], [v - a/2, v - a/2], color='lawngreen', linestyle='--')


def calculate_fI_curve(model, duration=1000, prerun_time=0, min_amp=0, max_amp=1, n=5, **kwargs):
    """
    Calculate the frequency-current (f-I) curve of the neuron model.

    Parameters
    ----------
    model : Model
        The neuron model.
    duration : int
        Duration of the simulation in ms.
    min_amp : float
        Minimum amplitude of the current injection in nA.
    max_amp : float
        Maximum amplitude of the current injection in nA.
    n : int
        Number of amplitudes to test.

    Returns
    -------
    dict
        A dictionary containing the current amplitudes, firing rates, and voltages.
    """

    seg = model.seg_tree.root
    duration = duration
    
    amps = np.round(np.linspace(min_amp, max_amp, n), 4)
    iclamp = model.iclamps[seg]
    rates = []
    vs = {}
    for amp in amps:
        iclamp.amp = amp
        model.run(duration=duration, prerun_time=prerun_time)
        spike_data = detect_somatic_spikes(model, **kwargs)
        n_spikes = len(spike_data['spike_times'])
        rate = n_spikes / iclamp.dur * 1000
        rates.append(rate)
        vs[amp] = model.simulator.recordings['v'][seg]

    return {
        'current_amplitudes': amps,
        'firing_rates': rates,
        'voltages': vs,
        'time': model.simulator.t
    }


def plot_fI_curve(data, ax=None, vshift=200, **kwargs):
    """
    Plot the f-I curve and somatic voltage traces.

    Parameters
    ----------
    data : dict
        A dictionary containing the current amplitudes, firing rates, and voltages.
        Can be obtained from `calculate_fI_curve`.
    ax : matplotlib.axes.Axes, optional
        The axes to plot on. If two axes are provided, the first will show the somatic voltage traces and the second will show the f-I curve.
        If a single axis is provided, it will show the f-I curve only.
        If None, a new figure will be created.
    vshift : int, optional
        The vertical shift for the somatic voltage traces. Default is 200.
    """

    if ax is None:
        _, ax = plt.subplots(1, 2, figsize=(5, 5))

    amps = data['current_amplitudes']
    rates = data['firing_rates']
    vs = data['voltages']
    t = data['time']

    if isinstance(ax, (list, np.ndarray)):        
        for i, (amp, v) in enumerate(vs.items()):
            ax[0].plot(t, np.array(v) - i*vshift, label=f'{amp} nA')
        ax[0].set_title('Somatic spikes')
        ax[0].legend()
        ax[0].axis('off')
        ax = ax[1]
    
    ax.plot(amps, rates, color='gray', zorder=0)
    for a, r in zip(amps, rates):
        ax.plot(a, r, 'o', zorder=1)
    ax.set_xlabel('Current (nA)')
    ax.set_ylabel('Firing rate (Hz)')
    ax.set_title('f-I curve')


# =============================================================================
# DENDRITIC PROPERTIES
# =============================================================================

def calculate_voltage_attenuation(model):
    """
    Calculate the voltage attenuation along the dendrites.

    Parameters
    ----------
    model : Model
        The neuron model.

    Returns
    -------
    dict
        A dictionary containing the path distances, minimum voltages, and voltage attenuations
    """

    # Assuming one stimulation site and multiple recording sites including the stimulated site
    stimulated_segs = list(model.iclamps.keys())
    if len(stimulated_segs) != 1:
        print("Only one stimulation site is supported")
        return None
    recorded_segs = list(model.recordings['v'].keys())
    if len(recorded_segs) < 2:
        print("At least two recording sites are required")
        return None

    print(f"Stimulating segment: {stimulated_segs[0]}")
    print(f"Recording segments: {recorded_segs}")

    stimulated_seg = stimulated_segs[0]

    iclamp = model.iclamps[stimulated_seg]
    amp = iclamp.amp

    if amp >= 0:
        print("Stimulus amplitude must be negative")
        return None

    path_distances = [seg.path_distance() for seg in recorded_segs]

    start_ts = int(iclamp.delay / model.simulator.dt)
    stop_ts = int((iclamp.delay + iclamp.dur) / model.simulator.dt)

    voltage_at_stimulated = np.array(model.simulator.recordings['v'][stimulated_seg])[start_ts:stop_ts]
    voltages = [np.array(model.simulator.recordings['v'][seg])[start_ts:stop_ts] for seg in recorded_segs]


    # Calculate voltage displacement from the resting potential
    delta_v_at_stimulated = voltage_at_stimulated[0] - voltage_at_stimulated[-2]# np.min(voltage_at_stimulated)
    delta_vs = [v[0] - v[-2] for v in voltages] # np.min(v) for v in voltages]

    min_voltages = [np.min(v) for v in voltages]
    end_voltages = [v[-2] for v in voltages]

    attenuation = [dv / delta_v_at_stimulated for dv in delta_vs]

    return {
        'path_distances': path_distances,
        'min_voltages': min_voltages,
        'end_voltages': end_voltages,
        'attenuation': attenuation
    }


def plot_voltage_attenuation(data, ax=None):

    path_distances = data['path_distances']
    attenuation = data['attenuation']

    if ax is None:
        _, ax = plt.subplots()

    ax.plot(path_distances, attenuation, 'o-')
    ax.set_ylim(-0.1, 1.1)
    ax.set_xlabel('Path distance (um)')
    ax.set_ylabel('Voltage attenuation')
    ax.set_title('Voltage attenuation')

def calculate_dendritic_nonlinearity(model, duration=1000, prerun_time=0, max_weight=None, n=None):
    """Calculate the expected and observed voltage changes for a range of synaptic weights.

    Parameters
    ----------
    model : Model
        The neuron model.
    duration : int
        Duration of the simulation in ms.
    max_weight : int
        Maximum synaptic weight to test.

    Returns
    -------
    dict
        A dictionary containing the expected and observed voltage changes.
    """

    recorded_segs = list(model.recordings['v'].keys())
    seg = recorded_segs[0]

    populations = list(model.populations.values())
    if len(populations) != 1:
        print("Only one population is supported")
        return None
    population = populations[0]
    if population.N != 1:
        print("Only one synapse should be placed on the dendrite")
        return None

    start_ts = int(population.input_params['start'] / model.simulator.dt)

    vs = {}
    delta_vs = []
    min_weight = 1
    if max_weight is None or min_weight is None or n is None:
        max_weight = population.input_params['weight']
        n = max_weight + 1

    weights = np.linspace(min_weight, max_weight, n, dtype=int)
    weights = np.unique(weights)

    for w in weights:
        population.update_input_params(weight=w)
        model.run(duration=duration, prerun_time=prerun_time)
        v = np.array(model.simulator.recordings['v'][seg])
        v_start = v[start_ts]
        v_max = np.max(v[start_ts:])
        delta_v = v_max - v_start
        delta_vs.append(delta_v)
        vs[w] = v
    unitary_delta_v = delta_vs[0]
    expected_delta_vs = [w * unitary_delta_v for w in weights]

    return {
        'expected_response': expected_delta_vs,
        'observed_response': delta_vs,
        'voltages': vs,
        'weights': weights,
        'time': model.simulator.t
    }


def plot_dendritic_nonlinearity(data, ax=None, vshift=200, **kwargs):
    """
    Plot the dendritic nonlinearity based on expected and observed voltage changes.

    Parameters
    ----------
    data : dict
        A dictionary containing the expected and observed voltage changes.
        Can be obtained from `calculate_dendritic_nonlinearity`.
    ax : matplotlib.axes.Axes, optional
        The axes to plot on. If two axes are provided, the first will show the voltage traces and the second will show the dendritic nonlinearity.
        If a single axis is provided, it will show the dendritic nonlinearity only.
        If None, a new figure will be created.
    vshift : int, optional
        The vertical shift for the voltage traces.
    """
    
    if ax is None:
        _, ax = plt.subplots(1, 2, figsize=(10, 5))

    expected_delta_vs = data['expected_response']
    delta_vs = data['observed_response']
    vs = data['voltages']
    t = data['time']

    if isinstance(ax, (list, np.ndarray)):        
        for i, (weight, v) in enumerate(vs.items()):
            ax[0].plot(t, np.array(v) - i*vshift, label=f'{weight} synapses')
        ax[0].set_title('Voltage traces')
        ax[0].legend()
        ax[0].axis('off')
        ax = ax[1]
    
    ax.plot(expected_delta_vs, delta_vs, zorder=1)
    ax.plot(expected_delta_vs, expected_delta_vs, color='gray', linestyle='--', zorder=0)
    for ep, ob in zip(expected_delta_vs, delta_vs):
        ax.plot(ep, ob, 'o', zorder=2)
    ax.set_xlabel('Expected voltage change (mV)')
    ax.set_ylabel('Observed voltage change (mV)')
    ax.set_title('Dendritic nonlinearity')
    

def calculate_sag_ratio(model):
    """
    Calculate the sag ratio of the neuron model.

    Parameters
    ----------
    model : Model
        The neuron model.

    Returns
    -------
    dict
        A dictionary containing the sag ratio and intermediate values.
    """
    v, t, dt, iclamp = get_somatic_data(model)

    start_ts = int(iclamp.delay / dt)
    stop_ts = int((iclamp.delay + iclamp.dur) / dt)
    min_ts = np.argmin(v[start_ts:stop_ts]) + start_ts
    v_min = np.min(v[start_ts: min_ts])

    a = v[stop_ts] - v_min
    b = v[start_ts] - v_min

    sag_ratio = a / b if b != 0 else np.nan

    print(f"Sag ratio: {a:.2f}/{b:.2f} = {sag_ratio:.2f}")
    return {
        'a': a,
        'b': b,
        'sag_ratio': sag_ratio,
    }




# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

class prerun:
    """
    Context manager to prerun a simulation for a specified duration.
    This is useful to stabilize the model before running the main simulation.
    """
    def __init__(self, model, duration=300, truncate=True):
        self.model = model
        if duration <= 0:
            raise ValueError("Duration must be a positive number.")
        self.duration = duration
        self.truncate = truncate
        self._original_iclamp_delays = {}
        self._original_input_params = {}

    def __enter__(self):
        
        self._original_iclamp_delays = {k: v.delay for k, v in self.model.iclamps.items()}
        self._original_input_params = {
            pop_name: (pop.input_params['start'], pop.input_params['end'])
            for pop_name, pop in self.model.populations.items()
        }
        for iclamp in self.model.iclamps.values():
            iclamp.delay += self.duration
        
        for pop in self.model.populations.values():
            start = pop.input_params['start'] + self.duration
            end = pop.input_params['end'] + self.duration
            pop.update_input_params(**{
                'start': start,
                'end': end
            })

        return self

    def __exit__(self, exc_type, exc_value, traceback):

        # Restore iClamp delays
        for seg, delay in self._original_iclamp_delays.items():
            self.model.iclamps[seg].delay = delay

        # Restore input timings
        for pop_name, (start, end) in self._original_input_params.items():
            self.model.populations[pop_name].update_input_params(**{'start': start, 'end': end})
    
        duration_timepoints = int(self.duration / self.model.simulator.dt)
        if len(self.model.simulator.t) > duration_timepoints and self.truncate:
            self._truncate()
        
    def _truncate(self):
        """Truncate the simulation time and recordings to the specified duration."""

        onset = int(self.duration / self.model.simulator.dt)

        self.model.simulator.t = self.model.simulator.t[onset:]
        self.model.simulator.t = [t - self.duration for t in self.model.simulator.t]

        for var, recs in self.model.recordings.items():
            for seg, rec in recs.items():
                recs[seg] = rec[onset:]

        self.model.simulator._duration -= self.duration


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

# Imports
from typing import List, Union, Callable
import os
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
from numpy import nan
import pandas as pd
import quantities as pq

# DendroTweaks imports
from dendrotweaks.simulators import NeuronSimulator
from dendrotweaks.biophys.io import MODFileLoader
from dendrotweaks.morphology import Domain
from dendrotweaks.biophys.groups import SegmentGroup
from dendrotweaks.biophys.distributions import Distribution
from dendrotweaks.path_manager import PathManager
import dendrotweaks.morphology.reduce as rdc
from dendrotweaks.utils import DOMAINS_TO_GROUPS
from dendrotweaks.utils import DEFAULT_FIT_MODELS

# Mixins
from dendrotweaks.model_io import IOMixin
from dendrotweaks.model_simulation import SimulationMixin

# Warnings configuration
import warnings

def custom_warning_formatter(message, category, filename, lineno, file=None, line=None):
    return f"WARNING: {message}\n({os.path.basename(filename)}, line {lineno})\n"

warnings.formatwarning = custom_warning_formatter



class Model(IOMixin, SimulationMixin):
    """
    A model object that represents a neuron model.

    The class incorporates various mixins to separate concerns while
    maintaining a flat interface.

    Parameters
    ----------
    name : str
        The name of the model.
    simulator_name : str
        The name of the simulator to use (either 'NEURON' or 'Jaxley').
    path_to_data : str
        The path to the data files where swc and mod files are stored.

    Attributes
    ----------
    path_to_model : str
        The path to the model directory.
    path_manager : PathManager
        The path manager for the model.
    mod_loader : MODFileLoader
        The MOD file loader.
    simulator_name : str
        The name of the simulator to use. Default is 'NEURON'.
    point_tree : PointTree
        The point tree representing the morphological reconstruction.
    sec_tree : SectionTree
        The section tree representing the morphology on the section level.
    mechanisms : dict
        A dictionary of mechanisms available for the model.
    domains_to_mechs : dict
        A dictionary mapping domains to mechanisms inserted in them.
    params : dict
        A dictionary mapping parameters to their distributions.
    d_lambda : float
        The spatial discretization parameter.
    seg_tree : SegmentTree
        The segment tree representing the morphology on the segment level.
    iclamps : dict
        A dictionary of current clamps in the model.
    populations : dict
        A dictionary of "virtual" populations forming synapses on the model.
    simulator : Simulator
        The simulator object to use.
    """

    def __init__(self, path_to_model,
                simulator_name='NEURON',) -> None:

        # Metadata
        self.path_to_model = path_to_model
        self._name = os.path.basename(os.path.normpath(path_to_model))
        self.morphology_name = ''
        self.version = ''
        self.path_manager = PathManager(path_to_model)
        self.simulator_name = simulator_name
        self._verbose = False

        # File managers
        self.mod_loader = MODFileLoader()

        # Morphology
        self.point_tree = None
        self.sec_tree = None
        self.domains = {}

        # Mechanisms
        self.mechanisms = {}
        self.domains_to_mechs = {}

        # Parameters
        self.params = {
            'cm': {'all': Distribution('constant', value=1)}, # uF/cm2
            'Ra': {'all': Distribution('constant', value=35.4)}, # Ohm cm
        }

        self.params_to_units = {
            'cm': pq.uF/pq.cm**2,
            'Ra': pq.ohm*pq.cm,
        }

        # Groups
        self._groups = []

        # Distributions
        # self.distributed_params = {}

        # Segmentation
        self.d_lambda = 0.1
        self.seg_tree = None

        # Stimuli
        self.iclamps = {}
        self.populations = {}

        # Simulator
        if simulator_name == 'NEURON':
            self.simulator = NeuronSimulator()
        elif simulator_name == 'Jaxley':
            self.simulator = JaxleySimulator()
        else:
            raise ValueError(
                'Simulator name not recognized. Use NEURON or Jaxley.')


    # -----------------------------------------------------------------------
    # PROPERTIES
    # -----------------------------------------------------------------------

    @property
    def name(self):
        """
        The name of the directory containing the model.
        """
        return self._name


    @property
    def verbose(self):
        """
        Whether to print verbose output.
        """
        return self._verbose


    @verbose.setter
    def verbose(self, value):
        self._verbose = value
        self.mod_loader.verbose = value


    @property
    def mechs_to_domains(self):
        """
        The dictionary mapping mechanisms to domains where they are inserted.
        """
        mechs_to_domains = defaultdict(set)
        for domain_name, mech_names in self.domains_to_mechs.items():
            for mech_name in mech_names:
                mechs_to_domains[mech_name].add(domain_name)
        return dict(mechs_to_domains)

    
    @property
    def groups(self):
        """
        The dictionary of segment groups in the model.
        """
        return {group.name: group for group in self._groups}


    @property
    def groups_to_parameters(self):
        """
        The dictionary mapping segment groups to parameters.
        """
        groups_to_parameters = {}
        for group in self._groups:
            groups_to_parameters[group.name] = {}
            for mech_name, params in self.mechs_to_params.items():
                if mech_name not in group.mechanisms:
                    continue
                groups_to_parameters[group.name] = params
        return groups_to_parameters


    @property
    def parameters_to_groups(self):
        """
        The dictionary mapping parameters to groups where they are distributed.
        """
        parameters_to_groups = defaultdict(list)
        for group in self._groups:
            for mech_name, params in self.mechs_to_params.items():
                if mech_name not in group.mechanisms:
                    continue
                for param in params:
                    parameters_to_groups[param].append(group.name)
        return dict(parameters_to_groups)


    @property
    def params_to_mechs(self):
        """
        The dictionary mapping parameters to mechanisms to which they belong.
        """
        params_to_mechs = {}
        # Sort mechanisms by length (longer first) to ensure specific matches
        sorted_mechs = sorted(self.mechanisms, key=len, reverse=True)
        for param in self.params:
            matched = False
            for mech in sorted_mechs:
                suffix = f"_{mech}"  # Define exact suffix
                if param.endswith(suffix):
                    params_to_mechs[param] = mech
                    matched = True
                    break
            if not matched:
                params_to_mechs[param] = "Independent"  # No match found
        return params_to_mechs


    @property
    def mechs_to_params(self):
        """
        The dictionary mapping mechanisms to parameters they contain.
        """
        mechs_to_params = defaultdict(list)
        for param, mech_name in self.params_to_mechs.items():
            mechs_to_params[mech_name].append(param)
        return dict(mechs_to_params)


    @property 
    def conductances(self):
        """
        A filtered dictionary of parameters that represent conductances.
        """
        return {param: value for param, value in self.params.items()
                if param.startswith('gbar')}


    @property
    def df_params(self):
        """
        A DataFrame of parameters and their distributions.
        """
        data = []
        for mech_name, params in self.mechs_to_params.items():
            for param in params:
                for group_name, distribution in self.params[param].items():
                    data.append({
                        'Mechanism': mech_name,
                        'Parameter': param,
                        'Group': group_name,
                        'Distribution': distribution if isinstance(distribution, str) else distribution.function_name,
                        'Distribution params': {} if isinstance(distribution, str) else distribution.parameters,
                    })
        df = pd.DataFrame(data)
        return df


    # ========================================================================
    # DOMAINS
    # ========================================================================
    def add_domain(self, name, type_idx, color, sections, distribute=True):
        """
        Adds a new empty domain to the model.

        Parameters
        ----------
        name : str
            The name of the domain.
        color : str
            The color assigned to the domain.
        type_idx : int
            The type index of the domain.
        sections : list[Section]
            The sections to include in the domain.
        distribute : bool, optional
            Whether to re-distribute the parameters after defining the domain. 
            Default is True.

        Notes
        -----
        This method does not automatically insert mechanisms into the newly 
        created domain. It is the user's responsibility to insert mechanisms 
        into the domain after its creation.

        Suggested type indices and colors:
                1: soma: orange
                2: axon: gold
                3: dend: forestgreen
                31: basal: seagreen
                4: apic: steelblue
                41: trunk: skyblue
                42: tuft: plum
                43: oblique: rosybrown
        """
        if name in self.domains:
            raise ValueError(f"Domain '{name}' already exists.")
        if not name or not name.strip():
            raise ValueError("Domain name cannot be empty.")
        if not sections:
            raise ValueError('No sections provided to define the domain.')
            
        complement_sections = set(self.sec_tree.sections) - set(sections)
        self._validate_domain_type_idx(complement_sections, type_idx)
        self._validate_domain_color(complement_sections, color)

        domain = Domain(name, type_idx, color)
        self._add_domain_groups(domain.name)
        self.domains[domain.name] = domain
        self.domains_to_mechs[domain.name] = set()

        self.extend_domain(name, sections, distribute=distribute)


    def _validate_domain_type_idx(self, complement_sections, type_idx):

        unique_complement_type_ids = set(sec.type_idx for sec in complement_sections)
        if type_idx in unique_complement_type_ids:
            raise ValueError(f'Type index {type_idx} is already used by another domain.')


    def _validate_domain_color(self, complement_sections, color):

        unique_complement_colors = set(sec.domain_color for sec in complement_sections)
        if color in unique_complement_colors:
            raise ValueError(f'Color {color} is already used by another domain.')
    

    def update_domain_name(self, old_name, new_name):
        """
        Update the name of a domain.

        Parameters
        ----------
        old_name : str
            The current name of the domain.
        new_name : str
            The new name to assign to the domain.
        """
        
        if new_name in self.domains:
            raise ValueError(f'Domain {new_name} already exists.')
        
        domain = self.domains[old_name]
        domain.name = new_name
        self.domains[domain.name] = domain
        self.domains.pop(old_name)
        # Update groups
        self._add_domain_groups(domain.name)
        self._remove_domain_groups(old_name)
        # Update domains_to_mechs
        self.domains_to_mechs[domain.name] = self.domains_to_mechs.pop(old_name)
        self._remove_empty()


    def update_domain_type_idx(self, name, new_type_idx):
        """
        Update the type index of a domain.

        Notes
        -----
        Suggested type indices:
                1: soma
                2: axon
                3: dend
                31: basal
                4: apic
                41: trunk
                42: tuft
                43: oblique
        """

        domain = self.domains[name]

        if new_type_idx is not None:
            if new_type_idx == domain.type_idx:
                return
            existing_type_ids = set(domain.type_idx for domain in self.domains.values()) - {domain.type_idx}
            if new_type_idx in existing_type_ids:
                raise ValueError(f'Type index {new_type_idx} is already used by another domain.')
            domain.type_idx = new_type_idx


    def update_domain_color(self, name, new_color, force=False):
        """
        Update the color of a domain.

        Notes
        -----
        Suggested colors:
                soma: orange
                axon: gold
                dend: forestgreen
                basal: seagreen
                apic: steelblue
                trunk: skyblue
                tuft: plum
                oblique: rosybrown
        """

        domain = self.domains[name]

        if new_color is not None:
            if new_color == domain.color:
                return
            existing_colors = set(domain.color for domain in self.domains.values()) - {domain.color}
            if new_color in existing_colors and not force:
                raise ValueError(f'Color {new_color} is already used by another domain.')
            domain.color = new_color


    def extend_domain(self, name, sections, distribute=True):
        """
        Extends an existing domain by adding sections to it.

        Parameters
        ----------
        name : str
            The name of the domain to extend.
        sections : list[Section]
            The sections to add to the domain.
        distribute : bool, optional
            Whether to re-distribute the parameters after extending the domain. 
            Default is True.

        Notes
        -----
        If the domain already exists and is being extended, 
        mechanisms will be inserted automatically 
        into the newly added sections.
        """

        domain = self.domains.get(name)
        if domain is None:
            raise ValueError(f'Domain {name} does not exist.')

        if not sections:
            raise ValueError('No sections provided to extend the domain.')

        # Find sections that are not in the domain yet
        sections_to_move = [sec for sec in sections 
                            if sec.domain_name != name]
        if not sections_to_move:
            warnings.warn(f'Sections already in domain {name}.')
            return

        # Remove sections from their old domains
        for sec in sections_to_move:
            old_domain = self.domains[sec.domain_name]
            old_domain.remove_section(sec)
            for mech_name in self.domains_to_mechs[old_domain.name]:
                # TODO: What if section is already in domain? Can't be as
                # we use a filtered list of sections.
                mech = self.mechanisms[mech_name]
                sec.uninsert_mechanism(mech)
            
        # Add sections to the new domain
        for sec in sections_to_move:
            domain.add_section(sec)
            # Important: here we insert mechanisms only if we extend the domain,
            # i.e. the domain already exists and has mechanisms.
            # If the domain is new, we DO NOT insert mechanisms automatically
            # and leave it to the user to do so.
            for mech_name in self.domains_to_mechs.get(domain.name, set()):
                mech = self.mechanisms[mech_name]
                sec.insert_mechanism(mech)

        self._remove_empty()
        self.sec_tree.sort(sort_children=True, force=True)

        if distribute:
            self.distribute_all()


    def _add_domain_groups(self, domain_name):
        """
        Manage groups when a domain is added.
        """
        # Add new domain to `all` group
        if self.groups.get('all'):
            self.groups['all'].domains.append(domain_name)
        # Create a new group for the domain
        group_name = DOMAINS_TO_GROUPS.get(domain_name, domain_name)
        self.add_group(group_name, domains=[domain_name])


    def _remove_domain_groups(self, domain_name):
        """
        Manage groups when a domain is removed.
        """
        for group in self._groups:
            if domain_name in group.domains:
                group.domains.remove(domain_name)
    

    def _remove_empty(self):
        self._remove_empty_domains()
        self._remove_uninserted_mechanisms()
        self._remove_empty_groups()


    def _remove_empty_domains(self):
        """
        """
        empty_domains = [domain for domain in self.domains.values() 
            if domain.is_empty()]
        for domain in empty_domains:
            warnings.warn(f'Domain {domain.name} is empty and will be removed.')
            self.domains.pop(domain.name)
            self.domains_to_mechs.pop(domain.name)
            self._remove_domain_groups(domain.name)


    def _remove_uninserted_mechanisms(self):
        mech_names = list(self.mechs_to_params.keys())
        mechs = [self.mechanisms[mech_name] for mech_name in mech_names
             if mech_name != 'Independent']
        uninserted_mechs = [mech for mech in mechs
                    if mech.name not in self.mechs_to_domains]
        for mech in uninserted_mechs:
            warnings.warn(f'Mechanism {mech.name} is not inserted in any domain and will be removed.')
            self._remove_mechanism_params(mech)


    def _remove_empty_groups(self):
        empty_groups = [group for group in self._groups 
                        if not any(seg in group 
                        for seg in self.seg_tree)]
        for group in empty_groups:
            warnings.warn(f'Group {group.name} is empty and will be removed.')
            self.remove_group(group.name)


    # ========================================================================
    # MECHANISMS
    # ========================================================================

    def insert_mechanism(self, mechanism_name: str, 
                         domain_name: str, distribute=True):
        """
        Insert a mechanism into all sections in a domain.

        Parameters
        ----------
        mechanism_name : str
            The name of the mechanism to insert.
        domain_name : str
            The name of the domain to insert the mechanism into.
        distribute : bool, optional
            Whether to distribute the parameters after inserting the mechanism.
        """
        mech = self.mechanisms[mechanism_name]
        domain = self.domains[domain_name]

        # domain.insert_mechanism(mech)
        self.domains_to_mechs[domain_name].add(mech.name)
        for sec in domain.sections:
            sec.insert_mechanism(mech)
        self._add_mechanism_params(mech)

        # TODO: Redistribute parameters if any group contains this domain
        if distribute:
            for param_name in self.params:
                self.distribute(param_name)
        

    def _add_mechanism_params(self, mech):
        """
        Update the parameters when a mechanism is inserted.
        By default each parameter is set to a constant value
        through the entire cell.
        """
        for param_name, value in mech.range_params_with_suffix.items():
            self.params[param_name] = {'all': Distribution('constant', value=value)}
        
        if hasattr(mech, 'ion') and mech.ion in ['na', 'k', 'ca']:
            self._add_equilibrium_potentials_on_mech_insert(mech.ion)


    def _add_equilibrium_potentials_on_mech_insert(self, ion: str) -> None:
        """
        """
        if ion == 'na' and not self.params.get('ena'):
            self.params['ena'] = {'all': Distribution('constant', value=50)}
        elif ion == 'k' and not self.params.get('ek'):
            self.params['ek'] = {'all': Distribution('constant', value=-77)}
        elif ion == 'ca' and not self.params.get('eca'):
            self.params['eca'] = {'all': Distribution('constant', value=140)}


    def uninsert_mechanism(self, mechanism_name: str, 
                            domain_name: str):
        """
        Uninsert a mechanism from all sections in a domain

        Parameters
        ----------
        mechanism_name : str
            The name of the mechanism to uninsert.
        domain_name : str
            The name of the domain to uninsert the mechanism from.
        """
        mech = self.mechanisms[mechanism_name]
        domain = self.domains[domain_name]

        # domain.uninsert_mechanism(mech)
        for sec in domain.sections:
            sec.uninsert_mechanism(mech)
        self.domains_to_mechs[domain_name].remove(mech.name)

        if not self.mechs_to_domains.get(mech.name):
            warnings.warn(f'Mechanism {mech.name} is not inserted in any domain and will be removed.')
            self._remove_mechanism_params(mech)

    
    def _remove_mechanism_params(self, mech):
        for param_name in self.mechs_to_params.get(mech.name, []):
            self.params.pop(param_name)

        if hasattr(mech, 'ion') and mech.ion in ['na', 'k', 'ca']:
            self._remove_equilibrium_potentials_on_mech_uninsert(mech.ion)


    def _remove_equilibrium_potentials_on_mech_uninsert(self, ion: str) -> None:
        """
        """
        for mech_name, mech in self.mechanisms.items():
            if hasattr(mech, 'ion'):
                if mech.ion == mech.ion: return

        if ion == 'na':
            self.params.pop('ena', None)
        elif ion == 'k':
            self.params.pop('ek', None)
        elif ion == 'ca':
            self.params.pop('eca', None)


    # ========================================================================
    # PARAMETERS
    # ========================================================================

    # -----------------------------------------------------------------------
    # SEGMENT GROUPS (Where)
    # -----------------------------------------------------------------------

    def add_group(self, name, domains, select_by=None, min_value=None, max_value=None):
        """
        Add a group of sections to the model.

        Parameters
        ----------
        name : str
            The name of the group.
        domains : list[str]
            The domains to include in the group.
        select_by : str, optional
            The parameter to select the sections by. Can be 'diam', 'distance', 'domain_distance'.
        min_value : float, optional
            The minimum value of the parameter.
        max_value : float, optional
            The maximum value of the
        """
        if self.verbose: print(f'Adding group {name}...')
        group = SegmentGroup(name, domains, select_by, min_value, max_value)
        self._groups.append(group)
        

    def remove_group(self, group_name):
        """
        Remove a group from the model.

        Parameters
        ----------
        group_name : str
            The name of the group to remove.
        """
        # Remove group from the list of groups
        self._groups = [group for group in self._groups 
                        if group.name != group_name]
        # Remove distributions that refer to this group
        for param_name, groups_to_distrs in self.params.items():
            groups_to_distrs.pop(group_name, None)


    def move_group_down(self, name):
        """
        Move a group down in the list of groups.

        Parameters
        ----------
        name : str
            The name of the group to move down.
        """
        idx = next(i for i, group in enumerate(self._groups) if group.name == name)
        if idx > 0:
            self._groups[idx-1], self._groups[idx] = self._groups[idx], self._groups[idx-1]
        for param_name in self.distributed_params:
            self.distribute(param_name)


    def move_group_up(self, name):
        """
        Move a group up in the list of groups.

        Parameters
        ----------
        name : str
            The name of the group to move up.
        """
        idx = next(i for i, group in enumerate(self._groups) if group.name == name)
        if idx < len(self._groups) - 1:
            self._groups[idx+1], self._groups[idx] = self._groups[idx], self._groups[idx+1]
        for param_name in self.distributed_params:
            self.distribute(param_name)


    # -----------------------------------------------------------------------
    # DISTRIBUTIONS (How)
    # -----------------------------------------------------------------------

    def set_param(self, param_name: str,
                        group_name: str = 'all',
                        distr_type: str = 'constant',
                        **distr_params):
        """
        Set a parameter for a group of segments.

        Parameters
        ----------
        param_name : str
            The name of the parameter to set.
        group_name : str, optional
            The name of the group to set the parameter for. Default is 'all'.
        distr_type : str, optional
            The type of the distribution to use. Default is 'constant'.
        distr_params : dict
            The parameters of the distribution.
        """

        if 'group' in distr_params:
            raise ValueError("Did you mean 'group_name' instead of 'group'?")

        if param_name in ['temperature', 'v_init']:
            setattr(self.simulator, param_name, distr_params['value'])
            return

        for key, value in distr_params.items():
            if not isinstance(value, (int, float)) or value is nan:
                raise ValueError(f"Parameter '{key}' must be a numeric value and not NaN, got {type(value).__name__} instead.")

        self.set_distribution(param_name, group_name, distr_type, **distr_params)
        self.distribute(param_name)


    def set_distribution(self, param_name: str,
                         group_name: None,
                         distr_type: str = 'constant',
                         **distr_params):
        """
        Set a distribution for a parameter.

        Parameters
        ----------
        param_name : str
            The name of the parameter to set.
        group_name : str, optional
            The name of the group to set the parameter for. Default is 'all'.
        distr_type : str, optional
            The type of the distribution to use. Default is 'constant'.
        distr_params : dict
            The parameters of the distribution.
        """
        
        if distr_type == 'inherit':
            distribution = 'inherit'
        else:
            distribution = Distribution(distr_type, **distr_params)
        self.params[param_name][group_name] = distribution


    def distribute_all(self):
        """
        Distribute all parameters to the segments.
        """
        groups_to_segments = {group.name: [seg for seg in self.seg_tree if seg in group] 
                         for group in self._groups}
        for param_name in self.params:
            self.distribute(param_name, groups_to_segments)

    
    def distribute(self, param_name: str, precomputed_groups=None):
        """
        Distribute a parameter to the segments.

        Parameters
        ----------
        param_name : str
            The name of the parameter to distribute.
        precomputed_groups : dict, optional
            A dictionary mapping group names to segments. Default is None.
        """
        if param_name == 'Ra':
            self._distribute_Ra(precomputed_groups)
            return

        groups_to_segments = precomputed_groups
        if groups_to_segments is None:
            groups_to_segments = {group.name: [seg for seg in self.seg_tree if seg in group] 
                                for group in self._groups}

        param_distributions = self.params[param_name]

        for group_name, distribution in param_distributions.items():
            
            filtered_segments = groups_to_segments[group_name]

            if distribution == 'inherit':
                for seg in filtered_segments:
                    value = seg.parent.get_param_value(param_name)
                    seg.set_param_value(param_name, value)
            else:
                for seg in filtered_segments:
                    value = distribution(seg.path_distance())
                    seg.set_param_value(param_name, value)


    def _distribute_Ra(self, precomputed_groups=None):
        """
        Distribute the axial resistance to the segments.
        """

        groups_to_segments = precomputed_groups
        if groups_to_segments is None:
            groups_to_segments = {group.name: [seg for seg in self.seg_tree if seg in group] 
                                for group in self._groups}

        param_distributions = self.params['Ra']

        for group_name, distribution in param_distributions.items():
            
            filtered_segments = groups_to_segments[group_name]
            if distribution == 'inherit':
                raise NotImplementedError("Inheritance of Ra is not implemented.")
            else:
                for seg in filtered_segments:
                    value = distribution(seg._section.path_distance(0.5))
                    seg._section._ref.Ra = value


    def remove_distribution(self, param_name, group_name):
        """
        Remove a distribution for a parameter.

        Parameters
        ----------
        param_name : str
            The name of the parameter to remove the distribution for.
        group_name : str
            The name of the group to remove the distribution for.
        """
        self.params[param_name].pop(group_name, None)
        self.distribute(param_name)


    # -----------------------------------------------------------------------
    # FITTING
    # -----------------------------------------------------------------------

    def fit_distribution(self, param_name: str, segments, candidate_models=None, plot=True):
        if candidate_models is None:
            candidate_models = DEFAULT_FIT_MODELS

        from dendrotweaks.utils import mse

        values = [seg.get_param_value(param_name) for seg in segments]
        if all(np.isnan(values)):
            return None

        distances = [seg.path_distance() for seg in segments]
        distances, values = zip(*sorted(zip(distances, values)))

        best_score = float('inf')
        best_model = None
        best_params = None
        best_pred = None

        results = []

        for name, model in candidate_models.items():
            try:
                params, pred_values = model['fit'](distances, values)
                score = model.get('score', mse)(values, pred_values)
                complexity = model.get('complexity', 1)(params)
                results.append((name, score, params, complexity, pred_values))
            except Exception as e:
                warnings.warn(f"Model {name} failed to fit: {e}")

        # Sort results by score and complexity
        results.sort(key=lambda x: (np.round(x[1], 10), x[3]))

        best_model, best_score, best_params, _, best_pred = results[0]

        if plot:
            self.plot_param(param_name, show_nan=False)
            plt.plot(distances, best_pred, label=f'Best Fit: {best_model}', color='red', linestyle='--')
            plt.legend()

        return {'model': best_model, 'params': best_params, 'score': best_score}


    def _set_distribution(self, param_name, group_name, fit_result, plot=False):
        if fit_result is None:
            warnings.warn(f"No valid fit found for parameter {param_name}. Skipping distribution assignment.")
            return

        model_type = fit_result['model']
        params = fit_result['params']

        if model_type == 'poly':
            coeffs = np.array(params)
            coeffs = np.where(np.round(coeffs) == 0, coeffs, np.round(coeffs, 10))
            if len(coeffs) == 1:
                self.params[param_name][group_name] = Distribution('constant', value=coeffs[0])
            elif len(coeffs) == 2:
                self.params[param_name][group_name] = Distribution('linear', slope=coeffs[0], intercept=coeffs[1])
            else:
                self.params[param_name][group_name] = Distribution('polynomial', coeffs=coeffs.tolist())

        elif model_type == 'step':
            start, end, min_value, max_value = params
            self.params[param_name][group_name] = Distribution('step', max_value=max_value, min_value=min_value, start=start, end=end)


    # -----------------------------------------------------------------------
    # PLOTTING
    # -----------------------------------------------------------------------

    def plot_param(self, param_name, ax=None, show_nan=True):
        """
        Plot the distribution of a parameter in the model.

        Parameters
        ----------
        param_name : str
            The name of the parameter to plot.
        ax : matplotlib.axes.Axes, optional
            The axes to plot on. Default is None.
        show_nan : bool, optional
            Whether to show NaN values. Default is True.            
        """

        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 2))

        if param_name not in self.params:
            warnings.warn(f'Parameter {param_name} not found.')

        values = [(seg.path_distance(), seg.get_param_value(param_name)) for seg in self.seg_tree]
        colors = [seg.domain_color for seg in self.seg_tree]

        valid_values = [(x, y) for (x, y), color in zip(values, colors) if not pd.isna(y) and y != 0]
        zero_values = [(x, y) for (x, y), color in zip(values, colors) if y == 0]
        nan_values = [(x, 0) for (x, y), color in zip(values, colors) if pd.isna(y)]
        valid_colors = [color for (x, y), color in zip(values, colors) if not pd.isna(y) and y != 0]
        zero_colors = [color for (x, y), color in zip(values, colors) if y == 0]
        nan_colors = [color for (x, y), color in zip(values, colors) if pd.isna(y)]

        if valid_values:
            ax.scatter(*zip(*valid_values), c=valid_colors)
        if zero_values:
            ax.scatter(*zip(*zero_values), edgecolors=zero_colors, facecolors='none', marker='.')
        if nan_values and show_nan:
            ax.scatter(*zip(*nan_values), c=nan_colors, marker='x', alpha=0.5, zorder=0)
        ax.axhline(y=0, color='k', linestyle='--')

        ax.set_xlabel('Path distance')
        ax.set_ylabel(param_name)
        ax.set_title(f'{param_name} distribution')


    # ========================================================================
    # MORPHOLOGY
    # ========================================================================

    def get_sections(self, filter_function):
        """Filter sections using a lambda function.
        
        Parameters
        ----------
        filter_function : Callable
            The lambda function to filter sections.
        """
        return [sec for sec in self.sec_tree.sections if filter_function(sec)]


    def get_segments(self, group_names=None):
        """
        Get the segments in specified groups.

        Parameters
        ----------
        group_names : List[str]
            The names of the groups to get segments from.
        """
        if not isinstance(group_names, list):
            raise ValueError('Group names must be a list.')
        return [seg for group_name in group_names for seg in self.seg_tree.segments if seg in self.groups[group_name]]


    def remove_subtree(self, section):
        """
        Remove a subtree from the model.

        Parameters
        ----------
        section : Section
            The root section of the subtree to remove.
        """
        for domain in self.domains.values():
            for sec in section.subtree:
                if sec in domain.sections:
                    domain.remove_section(sec)
        self.sec_tree.remove_subtree(section)
        self._remove_empty()


    def merge_domains(self, domain_names: List[str]):
        """
        Merge two domains into one.
        """
        domains = [self.domains[domain_name] for domain_name in domain_names]
        for domain in domains[1:]:
            domains[0].merge(domain)
        self.remove_empty()


    def reduce_subtree(self, root, reduction_frequency=0, total_segments_manual=-1, fit=True):
        """
        Reduce a subtree to a single section.

        Parameters
        ----------
        root : Section
            The root section of the subtree to reduce.
        reduction_frequency : float, optional
            The frequency of the reduction. Default is 0.
        total_segments_manual : int, optional
            The number of segments in the reduced subtree. Default is -1 (automatic).
        fit : bool, optional
            Whether to create distributions for the reduced subtree by fitting
            the calculated average values. Default is True.
        """

        domain_name = root.domain_name
        parent = root.parent
        domains_in_subtree = [self.domains[domain_name] 
            for domain_name in set([sec.domain_name for sec in root.subtree])]
        if len(domains_in_subtree) > 1:
            # ensure the domains have the same mechanisms using self.domains_to_mechs
            domains_to_mechs = {domain_name: mech_names for domain_name, mech_names
                in self.domains_to_mechs.items() if domain_name in [domain.name for domain in domains_in_subtree]}
            common_mechs = set.intersection(*domains_to_mechs.values())
            if not all(mech_names == common_mechs
                    for mech_names in domains_to_mechs.values()):
                raise ValueError(
                    'The domains in the subtree have different mechanisms. '
                    'Please ensure that all domains in the subtree have the same mechanisms. '
                    'You may need to insert the missing mechanisms and set their conductances to 0 where they are not needed.'
                )
        elif len(domains_in_subtree) == 1:
            common_mechs = self.domains_to_mechs[domain_name].copy()
        
        inserted_mechs = {mech_name: mech for mech_name, mech
            in self.mechanisms.items()
            if mech_name in self.domains_to_mechs[domain_name]
        }

        subtree_without_root = [sec for sec in root.subtree if sec is not root]

        # Map original segment names to their parameters
        segs_to_params = rdc.map_segs_to_params(root, inserted_mechs)
        

        # Temporarily remove active mechanisms        
        for mech_name in inserted_mechs:
            if mech_name == 'Leak':
                continue
            for sec in root.subtree:
                mech = self.mechanisms[mech_name]
                sec.uninsert_mechanism(mech)

        # Disconnect
        root.disconnect_from_parent()

         # Calculate new properties of a reduced subtree
        new_cable_properties = rdc.get_unique_cable_properties(root._ref, reduction_frequency)
        new_nseg = rdc.calculate_nsegs(new_cable_properties, total_segments_manual)
        print(new_cable_properties)
        

         # Map segment names to their new locations in the reduced cylinder
        segs_to_locs = rdc.map_segs_to_locs(root, reduction_frequency, new_cable_properties)
        

        # Reconnect
        root.connect_to_parent(parent)

        # Delete the original subtree
        children = root.children[:]
        for child_sec in children:
            self.remove_subtree(child_sec)

        # Set passive mechanisms for the reduced cylinder:
        rdc.apply_params_to_section(root, new_cable_properties, new_nseg)
        

        # Reinsert active mechanisms
        for mech_name in inserted_mechs:
            if mech_name == 'Leak':
                continue
            for sec in root.subtree:
                mech = self.mechanisms[mech_name]
                sec.insert_mechanism(mech)
        
        # Replace locs with corresponding segs
        
        segs_to_reduced_segs = rdc.map_segs_to_reduced_segs(segs_to_locs, root)

        # Map reduced segments to lists of parameters of corresponding original segments
        reduced_segs_to_params = rdc.map_reduced_segs_to_params(segs_to_reduced_segs, segs_to_params)
        
        # Set new values of parameters
        rdc.set_avg_params_to_reduced_segs(reduced_segs_to_params)
        rdc.interpolate_missing_values(reduced_segs_to_params, root)


        data = {
            'segs_to_params': segs_to_params,
            'segs_to_locs': segs_to_locs,
            'segs_to_reduced_segs': segs_to_reduced_segs,
            'reduced_segs_to_params': reduced_segs_to_params,
        }

        if not fit:
            return data

        root_segs = [seg for seg in root.segments]
        params_to_fits = {}
        # for param_name in self.params:
        common_mechs.add('Independent')
        for mech in common_mechs:
            for param_name in self.mechs_to_params[mech]:
                fit_result = self.fit_distribution(param_name, segments=root_segs, plot=False)
                params_to_fits[param_name] = fit_result
        
        
        # Create new domain
        reduced_domains = [domain_name for domain_name in self.domains if domain_name.startswith('reduced')]
        new_reduced_domain_name = f'reduced_8{len(reduced_domains)}'
        new_reduced_domain_type_idx = int(f'8{len(reduced_domains)}')
        group_name = new_reduced_domain_name

        old_domain = root.domain_name
        self.update_domain_name(old_domain, new_reduced_domain_name)
        self.update_domain_type_idx(new_reduced_domain_name, new_reduced_domain_type_idx)
        self.update_domain_color(new_reduced_domain_name, 'palevioletred')
            
        # # Fit distributions to data for the group
        for param_name, fit_result in params_to_fits.items():
            self._set_distribution(param_name, group_name, fit_result, plot=True)

        # # Distribute parameters
        self.distribute_all()

        data.update({
            'params_to_fits': params_to_fits,
            'domain_name': new_reduced_domain_name, 
            'group_name': group_name,})
        return data


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

from dendrotweaks.morphology.io.reader import SWCReader
from dendrotweaks.morphology.io.validation import validate_tree
from dendrotweaks.morphology.io.factories import create_point_tree
from dendrotweaks.morphology.io.factories import create_section_tree
from dendrotweaks.morphology.io.factories import create_segment_tree
from dendrotweaks.morphology.io.factories import create_domains


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

import pandas as pd

SWC_IDS_TO_DOMAINS = {
    0: 'undefined',
    1: 'soma',
    2: 'axon',
    3: 'dend',
    31: 'basal',
    4: 'apic',
    41: 'trunk',
    42: 'tuft',
    43: 'oblique',
    6: 'neurite',
    7: 'glia',
    8: 'reduced',
}

DOMAINS_TO_COLORS = {
    'undefined': '#7F7F7F',
    'soma': '#E69F00',
    'axon': '#F0E442',
    'dend': '#019E73',
    'basal': '#31A354',
    'apic': '#0072B2',
    'trunk': '#56B4E9',
    'tuft': '#A55194',
    'oblique': '#8C564B',
    'neurite': '#D55E00',
    'glia': '#D62728',
    'reduced': '#E377C2',
}

DEFAULT_IDS = list(SWC_IDS_TO_DOMAINS.keys())

class SWCReader():
    """
    Reads an SWC file and returns a DataFrame.
    """

    def __init__(self):
        pass


    def read_file(self, path_to_swc_file: str) -> pd.DataFrame:
        """
        Read the SWC file and return a DataFrame enriched with domain & color metadata.
        """
        with open(path_to_swc_file, 'r') as f:
            raw_lines = f.readlines()

        # 1. Clean up lines
        lines = self._clean_lines(raw_lines)

        # 2. Read SWC table
        df = self._read_swc_table(path_to_swc_file)
        unique_type_ids = df['Type'].unique()

        # 3. Extract data from header (if available)
        domain_map, color_map = self._build_mappings(lines, unique_type_ids)

        # 4. Add domain and color columns
        df.insert(df.columns.get_loc("Type") + 1, "Domain", df["Type"].map(domain_map))
        df.insert(df.columns.get_loc("Domain") + 1, "Color", df["Domain"].map(color_map))

        return df
        

    @staticmethod
    def rename_domain(df, type_idx: int, new_domain_name: str, new_color: str):
        """
        Rename a domain in the DataFrame by changing the domain name and color for a given type index.
        """
        df.loc[df['Type'] == type_idx, 'Domain'] = new_domain_name
        df.loc[df['Type'] == type_idx, 'Color'] = new_color


    @staticmethod
    def replace_domain(df, old_type_idx: int, new_type_idx: int):
        """
        Replace a domain in the DataFrame by changing the type index, domain name, and color.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame containing the SWC data (generated by read_file).
        old_type_idx : int
            The type index of the domain to be replaced.
        new_type_idx : int
            The type index of the new domain.
        """
        new_domain_name = df.loc[df['Type'] == new_type_idx, 'Domain'].iloc[0]
        new_color = df.loc[df['Type'] == new_type_idx, 'Color'].iloc[0]
        df.loc[df['Type'] == old_type_idx, 'Type'] = new_type_idx
        df.loc[df['Type'] == new_type_idx, 'Domain'] = new_domain_name
        df.loc[df['Type'] == new_type_idx, 'Color'] = new_color

    @staticmethod
    def remove_domain(df, type_idx: int):
        """
        Remove a domain from the DataFrame by dropping all rows with the given type index.
        """
        df.drop(df[df['Type'] == type_idx].index, inplace=True)


    @staticmethod
    def _clean_lines(lines):
        """Strip whitespace, collapse multiple spaces, drop empty lines."""
        return [' '.join(line.split()) for line in lines if line.strip()]


    @staticmethod
    def _read_swc_header(lines):
        """
        Extract domain and color mappings from SWC header.
        """
        domain_map = {}
        color_map = {}

        domain_line = next((l for l in lines if l.startswith("# DOMAIN_NAMES")), None)
        color_line  = next((l for l in lines if l.startswith("# DOMAIN_COLORS")), None)

        if domain_line:
            items = domain_line.replace("# DOMAIN_NAMES", "").strip().split()
            for item in items:
                t, dom = item.split(":", 1)
                domain_map[int(t)] = dom

        if color_line:
            items = color_line.replace("# DOMAIN_COLORS", "").strip().split()
            for item in items:
                t, col = item.split(":", 1)
                domain_name = domain_map.get(int(t))
                color_map[domain_name] = col

        return domain_map, color_map


    def _build_mappings(self, lines, unique_type_ids):
        """
        Build domain and color mappings from SWC header or use defaults.
        """
        domain_map = SWC_IDS_TO_DOMAINS.copy()
        color_map = DOMAINS_TO_COLORS.copy()
        
        # Extract and apply header mappings
        header_domains, header_colors = self._read_swc_header(lines)
        domain_map.update(header_domains)
        color_map.update(header_colors)
        
        # Handle missing type IDs (reduced and custom domains)
        for t in unique_type_ids:
            if t in domain_map:
                continue
            domain_map[t] = f'custom_{t}'
            color_map[f'custom_{t}'] = '#17BECF'
        
        return domain_map, color_map


    @staticmethod
    def _read_swc_table(path):
        """Read the numeric part of an SWC file into a DataFrame."""
        df = pd.read_csv(
            path,
            sep=' ',
            header=None,
            comment='#',
            names=['Index', 'Type', 'X', 'Y', 'Z', 'R', 'Parent'],
            index_col=False,
            dtype={
                'Index': int,
                'Type': int,
                'X': float,
                'Y': float,
                'Z': float,
                'R': float,
                'Parent': int
            }
        )

        # Fix zero-radii case:
        if (df['R'] == 0).all():
            df['R'] = 1.0

        if df['Index'].duplicated().any():
            raise ValueError("The SWC file contains duplicate node ids.")

        return df


    @staticmethod
    def plot_raw_data(df, ax, **kwargs):
        """
        Plot the raw data from the SWC file using the DataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame containing the SWC data (generated by read_file).
        ax : matplotlib.pyplot.Axes
            The axes to plot on.
        """
        for t in df['Type'].unique():
            mask = df['Type'] == t
            color = df[mask]['Color'].iloc[0]
            if pd.isna(color):
                color = 'black'
            ax.scatter(xs=df[mask]['X'], ys=df[mask]['Y'], zs=df[mask]['Z'], 
                    c=color, label=f'Type {t}', **kwargs)
        ax.legend()


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

from dendrotweaks.morphology.trees import Node, Tree
from dendrotweaks.morphology.point_trees import Point, PointTree
from dendrotweaks.morphology.sec_trees import NeuronSection, Section, SectionTree
from dendrotweaks.morphology.domains import Domain
from dendrotweaks.morphology.seg_trees import NeuronSegment, Segment, SegmentTree
from dendrotweaks.morphology.io.validation import validate_tree


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

from dendrotweaks.stimuli.iclamps import IClamp
from dendrotweaks.stimuli.synapses import Synapse
from dendrotweaks.stimuli.populations import Population


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

# Imports
import os
import json
from collections import defaultdict

import jinja2
import pandas as pd

# DendroTweaks imports
from dendrotweaks import __version__
from dendrotweaks.morphology.io import create_point_tree, create_section_tree, create_segment_tree
from dendrotweaks.morphology.io import create_domains
from dendrotweaks.biophys.io import create_channel, standardize_channel
from dendrotweaks.biophys.groups import SegmentGroup
from dendrotweaks.biophys.distributions import Distribution
from dendrotweaks.biophys.mechanisms import LeakChannel, CaDynamics
from dendrotweaks.biophys.mechanisms import FallbackChannel
from dendrotweaks.stimuli import Population
from dendrotweaks.utils import DOMAINS_TO_GROUPS, DOMAINS_TO_NEURON

# Warnings configuration
import warnings

def custom_warning_formatter(message, category, filename, lineno, file=None, line=None):
    return f"WARNING: {message}\n({os.path.basename(filename)}, line {lineno})\n"

warnings.formatwarning = custom_warning_formatter

class IOMixin():
    """
    Mixin for the Model class to handle file I/O operations.
    """

    # -----------------------------------------------------------------------
    # DIRECTORY MANAGEMENT
    # -----------------------------------------------------------------------

    def print_directory_tree(self, *args, **kwargs):
        """
        Print the directory tree.
        """
        return self.path_manager.print_directory_tree(*args, **kwargs)


    def list_morphologies(self, extension='swc'):
        """
        List the morphologies available for the model.
        """
        return self.path_manager.list_files('morphology', extension=extension)


    def list_biophys(self, extension='json'):
        """
        List the biophysical configurations available for the model.
        """
        return self.path_manager.list_files('biophys', extension=extension)


    def list_mechanisms(self, extension='mod'):
        """
        List the mechanisms available for the model.
        """
        return self.path_manager.list_files('mod', extension=extension)


    def list_stimuli(self, extension='json'):
        """
        List the stimuli configurations available for the model.
        """
        return self.path_manager.list_folders('stimuli')


    # -----------------------------------------------------------------------
    # MORPHOLOGY I/O
    # -----------------------------------------------------------------------

    def load_morphology(self, file_name, soma_notation='3PS', 
        align=True, sort_children=True, force=False) -> None:
        """
        Read an SWC file and build the SWC and section trees.

        Parameters
        ----------
        file_name : str
            The name of the SWC file to read.
        soma_notation : str, optional
            The notation of the soma in the SWC file. Can be '3PS' (three-point soma) or '1PS'. Default is '3PS'.
        align : bool, optional
            Whether to align the morphology to the soma center and align the apical dendrite (if present).
        sort_children : bool, optional
            Whether to sort the children of each node by increasing subtree size
            in the tree sorting algorithms. If True, the traversal visits 
            children with shorter subtrees first and assigns them lower indices. If False, children
            are visited in their original SWC file order (matching NEURON's behavior).
        """
        # self.name = file_name.split('.')[0]
        self.morphology_name = file_name.replace('.swc', '')
        path_to_swc_file = self.path_manager.get_abs_path(f'morphology/{file_name}.swc')
        point_tree = create_point_tree(path_to_swc_file)
        # point_tree.remove_overlaps()
        point_tree.change_soma_notation(soma_notation)
        point_tree.sort(sort_children=sort_children, force=force)
        if align:    
            point_tree.shift_coordinates_to_soma_center()
            point_tree.align_apical_dendrite()
            point_tree.round_coordinates(8)
        self.point_tree = point_tree

        sec_tree = create_section_tree(point_tree)
        sec_tree.sort(sort_children=sort_children, force=force)
        self.sec_tree = sec_tree

        self.domains = create_domains(sec_tree)

        self.create_and_reference_sections_in_simulator()
        seg_tree = create_segment_tree(sec_tree)
        self.seg_tree = seg_tree

        self._add_default_segment_groups()
        self._initialize_domains_to_mechs()

        d_lambda = self.d_lambda
        self.set_segmentation(d_lambda=d_lambda)
              

    def create_and_reference_sections_in_simulator(self):
        """
        Create and reference sections in the simulator.
        """
        if self.verbose: print(f'Building sections in {self.simulator_name}...')
        for sec in self.sec_tree.sections:
            sec.create_and_reference()
        n_sec = len([sec._ref for sec in self.sec_tree.sections 
                    if sec._ref is not None])
        if self.verbose: print(f'{n_sec} sections created.')


    def _add_default_segment_groups(self):
        self.add_group('all', list(self.domains.keys()))
        for domain_name in self.domains:
            group_name = DOMAINS_TO_GROUPS.get(domain_name, domain_name)
            self.add_group(group_name, [domain_name])


    def _initialize_domains_to_mechs(self):
        for domain_name in self.domains:
            # Only if haven't been defined for the previous morphology
            # TODO: Check that domains match
            if not domain_name in self.domains_to_mechs: 
                self.domains_to_mechs[domain_name] = set()
        for domain_name, mech_names in self.domains_to_mechs.items():
            for mech_name in mech_names:
                mech = self.mechanisms[mech_name]
                self.insert_mechanism(mech, domain_name)


    def export_morphology(self, file_name):
        """
        Write the SWC tree to an SWC file.

        Parameters
        ----------
        version : str, optional
            The version of the morphology appended to the morphology name.
        """
        if file_name.endswith('.swc'): file_name = file_name[:-4]
        path_to_file = self.path_manager.get_abs_path(f'morphology/{file_name}.swc')
        
        self.point_tree.to_swc(path_to_file)


    # =======================================================================
    # BIOPHYSICS I/O
    # ========================================================================

    # -----------------------------------------------------------------------
    # MECHANISMS
    # -----------------------------------------------------------------------

    def add_default_mechanisms(self, recompile=False):
        """
        Add default mechanisms to the model.

        Parameters
        ----------
        recompile : bool, optional
            Whether to recompile the mechanisms.
        """
        leak = LeakChannel()
        self.mechanisms[leak.name] = leak

        cadyn = CaDynamics()
        self.mechanisms[cadyn.name] = cadyn

        self.load_mechanisms('default_mod', recompile=recompile)


    def add_mechanisms(self, dir_name:str = 'mod', recompile=True) -> None:
        """
        Add a set of mechanisms from an archive to the model.

        Parameters
        ----------
        dir_name : str, optional
            The name of the archive to load mechanisms from. Default is 'mod'.
        recompile : bool, optional
            Whether to recompile the mechanisms.
        """
        # Create Mechanism objects and add them to the model
        for mechanism_name in self.path_manager.list_files(dir_name, extension='mod'):
            self.add_mechanism(mechanism_name, 
                               load=True, 
                               dir_name=dir_name, 
                               recompile=recompile)            


    def add_mechanism(self, mechanism_name: str, 
                      python_template_name: str = 'default',
                      load=True, dir_name: str = 'mod', recompile=True
                      ) -> None:
        """
        Create a Mechanism object from the MOD file (or LeakChannel).

        Parameters
        ----------
        mechanism_name : str
            The name of the mechanism to add.
        python_template_name : str, optional
            The name of the Python template to use. Default is 'default'.
        load : bool, optional
            Whether to load the mechanism using neuron.load_mechanisms.
        """
        paths = self.path_manager.get_channel_paths(
            mechanism_name, 
            python_template_name=python_template_name
        )
        try:
            mech = create_channel(**paths)
        except NotImplementedError as e:
            if "KINETIC" in str(e):
                warnings.warn(
                    f"Could not import the '{mechanism_name}' channel because it uses an unsupported KINETIC block."
                    " A minimal fallback channel will be created for simulation only, supporting only the 'gbar' parameter."
                )
                mech = FallbackChannel(mechanism_name)
            else:
                raise
        # Add the mechanism to the model
        self.mechanisms[mech.name] = mech
        # Update the global parameters

        if load:
            self.load_mechanism(mechanism_name, dir_name, recompile)        


    def load_mechanisms(self, dir_name: str = 'mod', recompile=True) -> None:
        """
        Load mechanisms from an archive.

        Parameters
        ----------
        dir_name : str, optional
            The name of the archive to load mechanisms from.
        recompile : bool, optional
            Whether to recompile the mechanisms.
        """
        mod_files = self.path_manager.list_files(dir_name, extension='mod')
        for mechanism_name in mod_files:
            self.load_mechanism(mechanism_name, dir_name, recompile)


    def load_mechanism(self, mechanism_name, dir_name='mod', recompile=False) -> None:
        """
        Load a mechanism from the specified archive.

        Parameters
        ----------
        mechanism_name : str
            The name of the mechanism to load.
        dir_name : str, optional
            The name of the directory to load the mechanism from. Default is 'mod'.
        recompile : bool, optional
            Whether to recompile the mechanism.
        """
        path_to_mod_file = self.path_manager.get_abs_path(
            f"{dir_name}/{mechanism_name}.mod"
        )
        self.mod_loader.load_mechanism(
            path_to_mod_file=path_to_mod_file, recompile=recompile
        )


    def standardize_channel(self, channel_name, 
        python_template_name=None, mod_template_name=None, remove_old=True):
        """
        Standardize a channel by creating a new channel with the same kinetic
        properties using the standard equations.

        Parameters
        ----------
        channel_name : str
            The name of the channel to standardize.
        python_template_name : str, optional
            The name of the Python template to use.
        mod_template_name : str, optional
            The name of the MOD template to use. 
        remove_old : bool, optional
            Whether to remove the old channel from the model. Default is True.
        """

        # Get data to transfer
        channel = self.mechanisms[channel_name]
        channel_domain_names = [domain_name for domain_name, mech_names 
            in self.domains_to_mechs.items() if channel_name in mech_names]
        gbar_name = f'gbar_{channel_name}'
        gbar_distributions = self.params[gbar_name]
        # Kinetic variables cannot be transferred

        # Uninsert the old channel
        for domain_name in self.domains:
            if channel_name in self.domains_to_mechs[domain_name]:
                self.uninsert_mechanism(channel_name, domain_name)

        # Remove the old channel
        if remove_old:
            self.mechanisms.pop(channel_name)
              
        # Create, add and load a new channel
        paths = self.path_manager.get_standard_channel_paths(
            channel_name, 
            mod_template_name=mod_template_name
        )
        standard_channel = standardize_channel(channel, **paths)
        
        self.mechanisms[standard_channel.name] = standard_channel
        self.load_mechanism(standard_channel.name, recompile=True)

        # Insert the new channel
        for domain_name in channel_domain_names:
            self.insert_mechanism(standard_channel.name, domain_name)

        # Transfer data
        gbar_name = f'gbar_{standard_channel.name}'
        for group_name, distribution in gbar_distributions.items():
            self.set_param(gbar_name, group_name, 
                distribution.function_name, **distribution.parameters)


    # -----------------------------------------------------------------------
    # CONFIGURATION
    # -----------------------------------------------------------------------

    def to_dict(self):
        """
        Return a dictionary representation of the model.

        Returns
        -------
        dict
            The dictionary representation of the model.
        """
        return {
            'metadata': {
                'name': self.name,
                'software': f'dendrotweaks v{__version__}',
                'comments': 'Automatically generated by DendroTweaks.',
            },
            'd_lambda': self.d_lambda,
            'domains': {domain: sorted(list(mechs)) for domain, mechs in self.domains_to_mechs.items()},
            'groups': [
            group.to_dict() for group in self._groups
            ],
            'params': {
            param_name: {
                group_name: distribution if isinstance(distribution, str) else distribution.to_dict()
                for group_name, distribution in distributions.items()
            }
            for param_name, distributions in self.params.items()
            },
        }


    def from_dict(self, data):
        """
        Load the model from a dictionary.

        Parameters
        ----------
        data : dict
            The dictionary representation of the model.
        """
        if not self.name == data['metadata']['name']:
            raise ValueError('Model name does not match the data.')

        self.d_lambda = data['d_lambda']

        # Domains and mechanisms
        self.domains_to_mechs = {
            domain: set(mechs) for domain, mechs in data['domains'].items()
        }
        if self.verbose: print('Inserting mechanisms...')
        for domain_name, mechs in self.domains_to_mechs.items():
            for mech_name in mechs:
                self.insert_mechanism(mech_name, domain_name, distribute=False)
        # print('Distributing parameters...')
        # self.distribute_all()

        # Groups
        if self.verbose: print('Adding groups...')
        self._groups = [SegmentGroup.from_dict(group) for group in data['groups']]

        if self.verbose: print('Distributing parameters...')
        # Parameters
        self.params = {
            param_name: {
                group_name: distribution if isinstance(distribution, str) else Distribution.from_dict(distribution)
                for group_name, distribution in distributions.items()
            }
            for param_name, distributions in data['params'].items()
        }

        if self.verbose: print('Setting segmentation...')
        if self.sec_tree is not None:
            d_lambda = self.d_lambda
            self.set_segmentation(d_lambda=d_lambda)
            

    def export_biophys(self, file_name, **kwargs):
        """
        Export the biophysical properties of the model to a JSON file.

        Parameters
        ----------
        file_name : str
            The name of the file to write to.
        **kwargs : dict
            Additional keyword arguments to pass to `json.dump`.
        """        
        if file_name.endswith('.json'): file_name = file_name[:-5]
        path_to_json = self.path_manager.get_abs_path(f'biophys/{file_name}.json', create_dirs=True)
        if not kwargs.get('indent'):
            kwargs['indent'] = 4

        data = self.to_dict()
        with open(path_to_json, 'w') as f:
            json.dump(data, f, **kwargs)


    def load_biophys(self, file_name, recompile=True):
        """
        Load the biophysical properties of the model from a JSON file.

        Parameters
        ----------
        file_name : str
            The name of the file to read from.
        recompile : bool, optional
            Whether to recompile the mechanisms after loading. Default is True.
        """
        self.add_default_mechanisms()
        

        path_to_json = self.path_manager.get_abs_path(f'biophys/{file_name}.json')

        with open(path_to_json, 'r') as f:
            data = json.load(f)

        for mech_name in {mech for mechs in data['domains'].values() for mech in mechs}:
            if mech_name in ['Leak', 'CaDyn', 'Independent']:
                continue
            self.add_mechanism(mech_name, dir_name='mod', recompile=recompile)            

        self.from_dict(data)


    # =======================================================================
    # STIMULI I/O
    # ========================================================================

    # -----------------------------------------------------------------------
    # EXPORT
    # -----------------------------------------------------------------------

    def stimuli_to_dict(self):
        """
        Convert the stimuli to a dictionary representation.

        Returns
        -------
        dict
            The dictionary representation of the stimuli.
        """
        return {
            'metadata': {
                'name': self.name,
                'software': f'dendrotweaks v{__version__}',
                'comments': 'Automatically generated by DendroTweaks.',
            },
            'simulation': {
                'd_lambda': self.d_lambda,
                **self.simulator.to_dict(),
            },
            'stimuli': {
                'populations': {
                    pop_name: {'idx': i, **pop.to_dict()}
                    for i, (pop_name, pop) in enumerate(self.populations.items())
                }
            },
        }


    def _write_csv(self, df, path_to_csv):
        """
        Write a DataFrame to a CSV file.
        """
        if 'idx' in df.columns:
            df['idx'] = df['idx'].astype(int)
        if 'sec_idx' in df.columns:
            df['sec_idx'] = df['sec_idx'].astype(int)
        df.to_csv(path_to_csv, index=False)


    def _recordings_to_csv(self, path_to_csv=None):
        """
        Write the recordings to a CSV file.
        """
        rec_data = {
            'sec_idx': [],
            'loc': [],
            'var': [],
        }

        for var, recs in self.simulator.recordings.items():
            rec_data['var'].extend([var] * len(recs))
            rec_data['sec_idx'].extend([seg._section.idx for seg in recs])
            rec_data['loc'].extend([seg.x for seg in recs])

        df = pd.DataFrame(rec_data)
        if df.empty: return
        if path_to_csv: self._write_csv(df, path_to_csv)
        return df


    def _iclamps_to_csv(self, path_to_csv=None):
        """
        Write the IClamp data to a CSV file.
        """
        iclamp_data = {
            'sec_idx': [seg._section.idx for seg in self.iclamps],
            'loc': [seg.x for seg in self.iclamps],
            'amp': [iclamp.amp for iclamp in self.iclamps.values()],
            'delay': [iclamp.delay for iclamp in self.iclamps.values()],
            'dur': [iclamp.dur for iclamp in self.iclamps.values()],
        }
        df = pd.DataFrame(iclamp_data)
        if df.empty: return
        if path_to_csv: self._write_csv(df, path_to_csv)
        return df
        

    def _synapses_to_csv(self, path_to_csv=None):
        """
        Write the synapse data to a CSV file.
        """
        synapses_data = {
            'sec_idx': [],
            'loc': [],
            'pop_idx': [],
        }

        for i, (pop_name, pop) in enumerate(self.populations.items()):
            pop_data = pop.to_csv()
            synapses_data['pop_idx'] += [i] * len(pop_data['name'])
            synapses_data['sec_idx'] += pop_data['sec_idx']
            synapses_data['loc'] += pop_data['loc']

        df = pd.DataFrame(synapses_data)
        if df.empty: return
        if path_to_csv: self._write_csv(df, path_to_csv)
        return df


    def export_stimuli(self, file_name, **kwargs):
        """
        Export the stimuli to a JSON and CSV file.

        Parameters
        ----------
        file_name : str
            The name of the file to write to.
        **kwargs : dict
            Additional keyword arguments to pass to `json.dump`.
        """
        path_to_json = self.path_manager.get_abs_path(
            f'stimuli/{file_name}/config.json', 
            create_dirs=True
        )

        data = self.stimuli_to_dict()

        if not kwargs.get('indent'):
            kwargs['indent'] = 4
        with open(path_to_json, 'w') as f:
            json.dump(data, f, **kwargs)

        path_to_recordings_csv = self.path_manager.get_abs_path(f'stimuli/{file_name}/recordings.csv')
        path_to_iclamps_csv = self.path_manager.get_abs_path(f'stimuli/{file_name}/iclamps.csv')
        path_to_synapses_csv = self.path_manager.get_abs_path(f'stimuli/{file_name}/synapses.csv')

        self._recordings_to_csv(path_to_recordings_csv)
        self._iclamps_to_csv(path_to_iclamps_csv)
        self._synapses_to_csv(path_to_synapses_csv)


    # -----------------------------------------------------------------------
    # IMPORT
    # -----------------------------------------------------------------------

    def _load_recordings(self, path_to_recordings_csv):
        """
        Load recordings from a CSV file.
        """
        df_recs = pd.read_csv(path_to_recordings_csv)

        for row in df_recs.itertuples(index=False):
            self.add_recording(
                self.sec_tree.sections[row.sec_idx], 
                row.loc, 
                var=row.var
            )


    def _load_iclamps(self, path_to_csv):
        """
        Load IClamps from a CSV file.
        """
        df_iclamps = pd.read_csv(path_to_csv)

        for row in df_iclamps.itertuples(index=False):
            self.add_iclamp(
                self.sec_tree.sections[row.sec_idx], 
                row.loc,
                row.amp,
                row.delay,
                row.dur
            )
    

    def _load_populations(self, path_to_csv, data):
        """
        Load populations from a CSV and JSON file.
        """
        df_all_syn = pd.read_csv(path_to_csv)

        for pop_name, pop_data in data['stimuli']['populations'].items():

            df_pop = df_all_syn[df_all_syn['pop_idx'] == pop_data['idx']]

            segments = [self.sec_tree.sections[sec_idx](loc) 
                        for sec_idx, loc in zip(df_pop['sec_idx'], df_pop['loc'])]
            
            pop = Population(name=pop_name, 
                            segments=segments, 
                            N=pop_data['N'], 
                            syn_type=pop_data['syn_type'])
            
            syn_locs = [
                (
                    self.sec_tree.sections[sec_idx], 
                    loc
                ) 
                for sec_idx, loc in zip(
                    df_pop['sec_idx'].tolist(), 
                    df_pop['loc'].tolist()
                )
            ]
            
            pop.allocate_synapses(syn_locs=syn_locs)
            pop.update_kinetic_params(**pop_data['kinetic_params'])
            pop.update_input_params(**pop_data['input_params'])
            self._add_population(pop)


    def load_stimuli(self, file_name, load_legacy=False):
        """
        Load the stimuli from a JSON file.

        Parameters
        ----------
        file_name : str
            The name of the file to read from.
        """
        if not load_legacy:
            try:
                self._load_stimuli(file_name)
            except Exception as e:
                raise RuntimeError(
                    f"An error occurred while loading stimuli '{file_name}': {e}\n"
                    "If these stimuli were saved using the legacy format (v0.4.6 or earlier), "
                    f"please call load_stimuli('{file_name}', load_legacy=True) to attempt automatic conversion."
                )
        else:
            warnings.warn(
                f"Attempting to load using the legacy v0.4.6 format. "
                "\n(Support for the legacy stimuli format is deprecated and will be removed in future releases.)"
            )
            from dendrotweaks.stimuli.legacy_io import convert_legacy_stimuli
            convert_legacy_stimuli(
                path_manager=self.path_manager, 
                protocol_name=file_name,
                morphology_name=self.morphology_name,
                d_lambda=self.d_lambda
                )
            try:
                self._load_stimuli(file_name)
            except Exception as e:
                raise RuntimeError(f"An error occurred while loading legacy stimuli '{file_name}': {e}\n")


    def _clear_and_setup_simulation(self, data):
        """Common setup for both formats."""
        self.simulator.from_dict(data['simulation'])
        self.remove_all_stimuli()
        self.remove_all_recordings()
        
        # TODO: Check for cases when nseg is changed manually
        d_lambda = data['simulation'].get('d_lambda')
        if d_lambda is not None and self.d_lambda != d_lambda:
            self.d_lambda = d_lambda
            self.set_segmentation(d_lambda=d_lambda)


    def _load_stimuli(self, file_name):
        """
        Load the stimuli from a JSON file.

        Parameters
        ----------
        file_name : str
            The name of the file to read from.
        """
        
        path_to_json = self.path_manager.get_abs_path(f'stimuli/{file_name}/config.json')
        path_to_recordings_csv = self.path_manager.get_abs_path(f'stimuli/{file_name}/recordings.csv')
        path_to_iclamps_csv = self.path_manager.get_abs_path(f'stimuli/{file_name}/iclamps.csv')
        path_to_synapses_csv = self.path_manager.get_abs_path(f'stimuli/{file_name}/synapses.csv')
        
        with open(path_to_json, 'r') as f:
            data = json.load(f)

        self._clear_and_setup_simulation(data)

        if os.path.exists(path_to_iclamps_csv):
            self._load_iclamps(path_to_iclamps_csv)
      
        if os.path.exists(path_to_synapses_csv):
            self._load_populations(path_to_synapses_csv, data)

        if os.path.exists(path_to_recordings_csv):
            self._load_recordings(path_to_recordings_csv)


    # ========================================================================
    # EXPORT TO PLAIN SIMULATOR CODE
    # ========================================================================

    def export_to_NEURON(self, file_name, include_kinetic_params=True):
        """
        Export the model to a python file with plain NEURON code to reproduce the model.

        Parameters
        ----------
        file_name : str
            The name of the file to write to.
        """

        params_to_valid_domains = get_params_to_valid_domains(self)
        params = self.params if include_kinetic_params else filter_params(self)
        path_to_template = self.path_manager.get_abs_path('templates/NEURON_template.py', create_dirs=True)

        
        output = render_template(path_to_template,
        {
            'param_dict': params,
            'groups_dict': self.groups,
            'params_to_mechs': self.params_to_mechs,
            'domains_to_mechs': {domain: sorted(list(mechs)) for domain, mechs in self.domains_to_mechs.items()},
            'iclamps': self.iclamps,
            'recordings': self.simulator.recordings,
            'params_to_valid_domains': params_to_valid_domains,
            'domains_to_NEURON': {domain: get_neuron_domain(domain) for domain in self.domains},
        })

        if not file_name.endswith('.py'):
            file_name += '.py'
        path_to_model = self.path_manager.path_to_model
        output_path = os.path.join(path_to_model, file_name)
        with open(output_path, 'w') as f:
            f.write(output)


    def export_to_Jaxley(self, file_name):
        raise NotImplementedError("Export to plain Jaxley code is not implemented yet.")


# =======================================================================
# HELPER FUNCTIONS
# =======================================================================

def filter_params(model):
    """
    Filter out kinetic parameters from the model.

    Parameters
    ----------
    model : Model
        The model to filter.

    Returns
    -------
    Model
        The model with kinetic parameters filtered out.
    """
    filtered_params = {
        param: {
            group_name: distribution 
            for group_name, distribution in distributions.items() 
            if param in list(model.conductances.keys()) + ['cm', 'Ra', 'ena', 'ek', 'eca']} 
            for param, distributions in model.params.items()}
    return filtered_params


def get_neuron_domain(domain_name):
    base_domain, _, idx = domain_name.partition('_')
    if base_domain in ['reduced', 'custom'] and idx.isdigit():
        return f'{DOMAINS_TO_NEURON[base_domain]}{idx}'
    return DOMAINS_TO_NEURON.get(base_domain, 'dend_0')

def render_template(path_to_template, context):
    """
    Render a Jinja2 template.

    Parameters
    ----------
    path_to_template : str
        The path to the Jinja2 template.
    context : dict
        The context to render the template with.
    """
    with open(path_to_template, 'r') as f:
        template = jinja2.Template(f.read())
    return template.render(context)


def get_params_to_valid_domains(model):
    
    params_to_valid_domains = defaultdict(dict)

    for param, mech in model.params_to_mechs.items():
        for group_name, distribution in model.params[param].items():
            group = model.groups[group_name]
            valid_domains = [get_neuron_domain(domain) for domain in group.domains if mech == 'Independent' or mech in model.domains_to_mechs[domain]]
            params_to_valid_domains[param][group_name] = valid_domains

    return dict(params_to_valid_domains)


import json
import os
import pandas as pd
import numpy as np
from dendrotweaks import __version__
from dendrotweaks.morphology.io.reader import SWCReader
from dendrotweaks.morphology.io.factories import create_point_tree, create_section_tree


# ============================================================================
# SECTION MAPPING
# ============================================================================

def _legacy_sort_key(node):
    """
    Sorting key function for legacy ordering.
    Legacy version sorted by (type_idx, bifurcation_count).
    """
    return (node.type_idx, sum(1 for n in node.subtree if len(n.children) > 1))


def _new_sort_key(node):
    """
    Sorting key function for new ordering.
    New version sorts only by bifurcation_count (removed type_idx).
    """
    return sum(1 for n in node.subtree if len(n.children) > 1)


def _apply_sort_to_tree(tree, sort_key_func):
    """
    Apply custom sorting to a tree's children.
    
    Parameters
    ----------
    tree : Tree
        The tree to sort.
    sort_key_func : callable
        Function that takes a node and returns a sort key.
    """
    for node in tree._nodes:
        node.children = sorted(node.children, key=sort_key_func)


def _create_tree_with_sort(df, use_legacy_sort=False):
    """
    Create and sort a tree with specified sorting behavior.
    
    Parameters
    ----------
    df : DataFrame
        The morphology data.
    use_legacy_sort : bool
        If True, use legacy sorting (type_idx, bifurcations).
        If False, use new sorting (bifurcations only).
    
    Returns
    -------
    SectionTree
        The sorted section tree.
    """
    point_tree = create_point_tree(df)
    point_tree.remove_overlaps()
    point_tree.extend_sections()
    
    # Apply appropriate sorting to point tree
    if use_legacy_sort:
        _apply_sort_to_tree(point_tree, _legacy_sort_key)
    else:
        _apply_sort_to_tree(point_tree, _new_sort_key)
    
    # Reassign indices after custom child sorting
    point_tree.sort(sort_children=False, force=True)
    
    sec_tree = create_section_tree(point_tree)
    
    # Apply appropriate sorting to section tree
    if use_legacy_sort:
        _apply_sort_to_tree(sec_tree, _legacy_sort_key)
    else:
        _apply_sort_to_tree(sec_tree, _new_sort_key)
    
    # Reassign indices after custom child sorting
    sec_tree.sort(sort_children=False, force=True)
    
    return sec_tree


def _sections_match(sec_old, sec_new):
    """Check if two sections have matching coordinates."""
    return all(
        np.allclose(pt_old.x, pt_new.x) and
        np.allclose(pt_old.y, pt_new.y) and
        np.allclose(pt_old.z, pt_new.z)
        for pt_old, pt_new in zip(sec_old.points, sec_new.points)
    )


def _map_sections(old_sec_tree, new_sec_tree):
    """Map section indices from old tree to new tree."""
    old_to_new_idx = {}
    
    for sec_old in old_sec_tree.sections:
        for sec_new in new_sec_tree.sections:
            if _sections_match(sec_old, sec_new):
                old_to_new_idx[sec_old.idx] = sec_new.idx
                break
    
    return old_to_new_idx


def _obtain_section_mapping(path_manager, morphology_name):
    """
    Obtain mapping from legacy section indices to new section indices.
    
    The mapping accounts for the difference in sorting behavior:
    - Legacy: sorts children by (type_idx, bifurcation_count)
    - New: sorts children by bifurcation_count only
    
    Parameters
    ----------
    path_manager : PathManager
        Path manager for file access.
    morphology_name : str
        Name of the morphology file.
    
    Returns
    -------
    dict
        Mapping from old section indices to new section indices.
    """
    reader = SWCReader()
    df = reader.read_file(path_manager.get_abs_path(f'morphology/{morphology_name}.swc'))
    
    # Create trees with different sorting strategies
    legacy_sec_tree = _create_tree_with_sort(df, use_legacy_sort=True)
    new_sec_tree = _create_tree_with_sort(df, use_legacy_sort=False)
    
    return _map_sections(legacy_sec_tree, new_sec_tree)


def _remap_csv_section_indices(path_manager, protocol_name, mapping):
    """
    Update sec_idx column in all CSV files based on provided mapping.
    
    Parameters
    ----------
    path_manager : PathManager
        Path manager for file access.
    protocol_name : str
        Name of the protocol folder.
    mapping : dict
        Mapping from old to new section indices.
    """
    protocol_path = path_manager.get_abs_path(f'stimuli/{protocol_name}')
    
    for filename in os.listdir(protocol_path):
        if not filename.endswith('.csv'):
            continue
        
        file_path = os.path.join(protocol_path, filename)
        df = pd.read_csv(file_path)
        
        if 'sec_idx' in df.columns:
            df['sec_idx'] = df['sec_idx'].map(mapping)
            df.to_csv(file_path, index=False)


# ============================================================================
# FORMAT CONVERSION
# ============================================================================

def _read_legacy_files(path_manager, protocol_name):
    """Read legacy JSON and CSV files."""
    path_to_json = path_manager.get_abs_path(f'stimuli/{protocol_name}.json')
    path_to_csv = path_manager.get_abs_path(f'stimuli/{protocol_name}.csv')
    
    with open(path_to_json, 'r') as f:
        legacy_data = json.load(f)
    
    df_stimuli = pd.read_csv(path_to_csv)
    
    return legacy_data, df_stimuli


def _build_metadata(model_name):
    """Create metadata dictionary for new format."""
    return {
        'name': model_name,
        'software': f'dendrotweaks v{__version__}',
        'comments': 'Converted from legacy format by DendroTweaks.',
    }


def _convert_populations(legacy_data, df_stimuli):
    """Convert population data from legacy to new format."""
    syn_types = ['AMPA', 'NMDA', 'AMPA_NMDA', 'GABAa']
    populations = {}
    
    pop_idx = 0
    for syn_type in syn_types:
        if syn_type not in legacy_data['stimuli']['populations']:
            continue
        
        for i, pop_data in enumerate(legacy_data['stimuli']['populations'][syn_type]):
            pop_name = f"{syn_type}_{i}"
            populations[pop_name] = {
                'idx': pop_idx,
                'N': pop_data['N'],
                'syn_type': syn_type,
                'kinetic_params': pop_data['kinetic_params'],
                'input_params': pop_data['input_params']
            }
            pop_idx += 1
    
    return populations


def _convert_recordings(legacy_data, df_stimuli):
    """Convert recordings from legacy to new format."""
    df_recs = df_stimuli[df_stimuli['type'] == 'rec'].reset_index(drop=True)
    
    if df_recs.empty:
        return None
    
    return pd.DataFrame({
        'sec_idx': df_recs['sec_idx'].astype(int),
        'loc': df_recs['loc'],
        'var': [legacy_data['stimuli']['recordings'][idx]['var'] 
                for idx in df_recs['idx']]
    })


def _convert_iclamps(legacy_data, df_stimuli):
    """Convert IClamps from legacy to new format."""
    df_iclamps = df_stimuli[df_stimuli['type'] == 'iclamp'].reset_index(drop=True)
    
    if df_iclamps.empty:
        return None
    
    iclamp_params = legacy_data['stimuli']['iclamps']
    
    return pd.DataFrame({
        'sec_idx': df_iclamps['sec_idx'].astype(int),
        'loc': df_iclamps['loc'],
        'amp': [iclamp_params[idx]['amp'] for idx in df_iclamps['idx']],
        'delay': [iclamp_params[idx]['delay'] for idx in df_iclamps['idx']],
        'dur': [iclamp_params[idx]['dur'] for idx in df_iclamps['idx']]
    })


def _convert_synapses(legacy_data, df_stimuli):
    """Convert synapses from legacy to new format."""
    syn_types = ['AMPA', 'NMDA', 'AMPA_NMDA', 'GABAa']
    synapses_data = {'sec_idx': [], 'loc': [], 'pop_idx': []}
    
    pop_idx = 0
    for syn_type in syn_types:
        if syn_type not in legacy_data['stimuli']['populations']:
            continue
        
        df_syn = df_stimuli[df_stimuli['type'] == syn_type]
        num_pops = len(legacy_data['stimuli']['populations'][syn_type])
        
        for i in range(num_pops):
            df_pop = df_syn[df_syn['idx'] == i]
            
            synapses_data['sec_idx'].extend(df_pop['sec_idx'].astype(int).tolist())
            synapses_data['loc'].extend(df_pop['loc'].tolist())
            synapses_data['pop_idx'].extend([pop_idx] * len(df_pop))
            
            pop_idx += 1
    
    # sort by sec_idx and loc and pop_idx
    if synapses_data['sec_idx']:
        sort_indices = np.lexsort((synapses_data['loc'], synapses_data['sec_idx'], synapses_data['pop_idx']))
        synapses_data = {
            key: [synapses_data[key][i] for i in sort_indices]
            for key in synapses_data
        }

    return pd.DataFrame(synapses_data) if synapses_data['sec_idx'] else None


def _write_json(data, path_manager, file_name):
    """Write JSON configuration file."""
    path = path_manager.get_abs_path(
        f'stimuli/{file_name}/config.json',
        create_dirs=True
    )
    with open(path, 'w') as f:
        json.dump(data, f, indent=4)
    return path


def _write_csv(df, path_manager, file_name, csv_type):
    """Write CSV file if dataframe is not None."""
    if df is None:
        return None
    
    path = path_manager.get_abs_path(f'stimuli/{file_name}/{csv_type}.csv')
    df.to_csv(path, index=False)
    return path


def _print_summary(protocol_name, new_file_name, json_path, populations, 
                   df_recordings, df_iclamps, df_synapses):
    """Print conversion summary."""
    print()
    print(f"Conversion complete: {protocol_name} -> {new_file_name}")
    print(f"  - Config: {json_path}")
    print(f"  - Populations: {len(populations)}")
    print(f"  - Recordings: {len(df_recordings) if df_recordings is not None else 0}")
    print(f"  - IClamps: {len(df_iclamps) if df_iclamps is not None else 0}")
    print(f"  - Synapses: {len(df_synapses) if df_synapses is not None else 0}")


def convert_legacy_stimuli(path_manager, protocol_name, morphology_name=None, 
                           new_file_name=None, d_lambda=0.1):
    """
    Convert stimuli from legacy JSON and CSV format to new format.
    
    Parameters
    ----------
    path_manager : PathManager
        The path manager instance for handling file paths.
    protocol_name : str
        The name of the legacy protocol file to read from.
    morphology_name : str, optional
        The name of the morphology file for section index remapping.
        If provided, section indices will be remapped from legacy to new format.
    new_file_name : str, optional
        The name for the new format files. If None, uses protocol_name.
    """
    new_file_name = new_file_name or protocol_name
    
    # Read legacy files
    legacy_data, df_stimuli = _read_legacy_files(path_manager, protocol_name)
    
    # Convert data structures
    populations = _convert_populations(legacy_data, df_stimuli)
    df_recordings = _convert_recordings(legacy_data, df_stimuli)
    df_iclamps = _convert_iclamps(legacy_data, df_stimuli)
    df_synapses = _convert_synapses(legacy_data, df_stimuli)
    
    model_name = legacy_data.get('metadata', {}).get('name', 'unknown')

    # Build new JSON structure
    new_data = {
        'metadata': _build_metadata(model_name),
        'simulation': legacy_data['simulation'],
        'stimuli': {'populations': populations}
    }
    new_data['simulation']['d_lambda'] = d_lambda
    
    # Write new files
    json_path = _write_json(new_data, path_manager, new_file_name)
    _write_csv(df_recordings, path_manager, new_file_name, 'recordings')
    _write_csv(df_iclamps, path_manager, new_file_name, 'iclamps')
    _write_csv(df_synapses, path_manager, new_file_name, 'synapses')
    
    # Remap section indices if morphology provided
    if morphology_name:
        print(f"Remapping section indices using {morphology_name}...")
        mapping = _obtain_section_mapping(path_manager, morphology_name)
        _remap_csv_section_indices(path_manager, new_file_name, mapping)
        print(f"  - Remapped {len(mapping)} sections")
    
    # Print summary
    _print_summary(protocol_name, new_file_name, json_path, populations,
                   df_recordings, df_iclamps, df_synapses)


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

import os
from typing import List, Dict
import shutil

class PathManager:
    """
    A manager class for handling file and directory paths related to models data.

    Parameters
    ----------
    path_to_model : str
        The path to the model directory.

    Attributes
    ----------
    path_to_model : str
        The path to the model directory.
    paths : Dict[str, str]
        A dictionary of paths for different file types.
    """
    def __init__(self, path_to_model: str):
        if not os.path.isdir(path_to_model):
            raise FileNotFoundError(f"Directory {path_to_model} does not exist.")
        self.path_to_model = path_to_model
        self.paths = {
            'default_mod': os.path.join(self.path_to_data, 'Default'),
            'templates': os.path.join(self.path_to_data, 'Templates'),
            'morphology': os.path.join(self.path_to_model, 'morphology'),
            'biophys': os.path.join(self.path_to_model, 'biophys'),
            'mod': os.path.join(self.path_to_model, 'biophys', 'mod'),
            'python': os.path.join(self.path_to_model, 'biophys', 'python'),
            'stimuli': os.path.join(self.path_to_model, 'stimuli'),
        }
        self._ensure_paths_exist()


    def _ensure_paths_exist(self):
        """
        Ensure all necessary paths exist.
        """
        os.makedirs(self.path_to_model, exist_ok=True)
        for path in self.paths.values():
            os.makedirs(path, exist_ok=True)
        # if empty, copy default mod files
        if not os.listdir(self.paths['default_mod']):
            self.copy_default_mod_files()
        if not os.listdir(self.paths['templates']):
            self.copy_template_files()


    @property
    def path_to_data(self):
        """
        The path to the data directory, which is always the parent directory of path_to_model.
        """
        return os.path.abspath(os.path.join(self.path_to_model, os.pardir))


    def __repr__(self):
        return f"PathManager({self.path_to_model})"


    def copy_default_mod_files(self):
        """
        Copy default mod files to the data directory.
        """
        DEFAULT_MOD_DIR = os.path.join(os.path.dirname(__file__), 'biophys', 'default_mod')
        for file_name in os.listdir(DEFAULT_MOD_DIR):
            source = os.path.join(DEFAULT_MOD_DIR, file_name)
            destination = os.path.join(self.paths['default_mod'], file_name)
            shutil.copyfile(source, destination)


    def copy_template_files(self):
        """
        Copy template files to the data directory.
        """
        TEMPLATES_DIR = os.path.join(os.path.dirname(__file__), 'biophys', 'default_templates')
        for file_name in os.listdir(TEMPLATES_DIR):
            source = os.path.join(TEMPLATES_DIR, file_name)
            destination = os.path.join(self.paths['templates'], file_name)
            shutil.copyfile(source, destination)


    def remove_folder(self, relative_path: str) -> None:
        """
        Remove a folder and all its contents.

        Parameters
        ----------
        relative_path : str
            The absolute path to the folder to be removed.
        """
        folder_path = self.get_abs_path(relative_path)
        if os.path.isdir(folder_path):
            shutil.rmtree(folder_path, ignore_errors=True)


    def _resolve_root(self, relative_path: str) -> tuple[str, str]:
        """
        Given a relative path like 'stimuli/depolarizing/protocol.json',
        resolve the absolute root and the remainder.
        """
        relative_path = os.path.normpath(relative_path.strip())
        parts = relative_path.split(os.sep)
        root_key = parts[0]

        base_root = self.paths.get(root_key)
        if base_root is None:
            raise KeyError(
                f"Unknown top-level folder '{root_key}'. "
                f"Known roots: {list(self.paths.keys())}"
            )

        remainder = os.path.join(*parts[1:]) if len(parts) > 1 else ""
        return base_root, remainder


    def get_abs_path(self, relative_path: str, create_dirs: bool = False) -> str:
        """
        Get the absolute path to a file or directory based on a relative path.

        Parameters
        ----------
        relative_path : str
            Path relative to one of the registered roots.
        create_dirs : bool, default False
            If True, create the parent directories if they don't exist.

        Returns
        -------
        str
            Absolute path in OS-native format.

        Examples
        --------
        >>> pm = PathManager('/path/to/model')
        >>> pm.get_abs_path('stimuli/depolarizing/protocol.json')
        '/path/to/model/stimuli/depolarizing/protocol.json'
        """
        base_root, remainder = self._resolve_root(relative_path)
        abs_path = os.path.join(base_root, remainder)

        if create_dirs and remainder:
            os.makedirs(os.path.dirname(abs_path), exist_ok=True)

        return abs_path

    def list_folders(self, relative_path: str) -> List[str]:
        """
        List all folders in a given directory.

        Parameters
        ----------
        relative_path : str
            Path relative to one of the registered roots.

        Returns
        -------
        List[str]
            A list of folder names.

        Examples
        --------
        >>> pm = PathManager('/path/to/model')
        >>> pm.list_folders('stimuli')
        ['depolarizing_current', 'hyperpolarizing_current']
        """
        abs_path = self.get_abs_path(relative_path)

        if not os.path.isdir(abs_path):
            return []

        return [f for f in os.listdir(abs_path) if os.path.isdir(os.path.join(abs_path, f))]


    def list_files(self, relative_path: str, extension: str | None = None) -> List[str]:
        """
        List all files in a given directory with an optional extension filter.
        If the extension is None, su
        
        Parameters
        ----------
        relative_path : str
            Path relative to one of the registered roots.
        extension : str
            The file extension to filter by (e.g., 'mod', 'swc').
        
        Returns
        -------
        List[str]
            A list of file names.
        """
        abs_path = self.get_abs_path(relative_path)
        if extension and not extension.startswith('.'): 
            extension = f".{extension}"
        if not os.path.isdir(abs_path):
            return []
        return [f.replace(extension, '') 
                for f in os.listdir(abs_path) if f.endswith(extension)]


    def list_morphologies(self, extension: str = '.swc') -> List[str]:
        """
        List all SWC files.
        
        Returns
        -------
        List[str]
            A list of SWC file names.
        """
        return self.list_files('morphology', extension=extension)


    def list_stimuli(self) -> List[str]:
        """
        List all JSON files.
        
        Returns
        -------
        List[str]
            A list of JSON file names.
        """
        return self.list_folders('stimuli')


    def list_biophys(self):
        """
        List all biophysics files.
        
        Returns
        -------
        List[str]
            A list of biophysics file names.
        """
        return self.list_files('biophys', extension='.json')


    def print_directory_tree(self, subfolder=None) -> None:
        """
        Print a directory tree for a given file type.
        
        Parameters
        ----------
        file_type : str
            The type of file (e.g., 'mod', 'swc').
        """
        base_path = self.paths.get('model') if not subfolder else self.paths.get(subfolder)
        if not base_path or not os.path.isdir(base_path):
            print(f"Directory for {file_type} does not exist.")
            return

        def print_tree(path, prefix=""):
            items = os.listdir(path)
            for idx, item in enumerate(sorted(items)):
                is_last = idx == len(items) - 1
                connector = "└──" if is_last else "├──"
                item_path = os.path.join(path, item)
                print(f"{prefix}{connector} {item}")
                if os.path.isdir(item_path) and not item.startswith('x86_64'):
                    extension = "│   " if not is_last else "    "
                    print_tree(item_path, prefix + extension)

        print_tree(base_path)


    def get_channel_paths(self, mechanism_name: str, 
                          python_template_name: str = None) -> Dict[str, str]:
        """
        Get all necessary paths for creating a channel.

        Parameters
        ----------
        mechanism_name : str
            The name of the mechanism.
        python_template_name : str, optional
            The name of the Python template file.

        Returns
        -------
        Dict[str, str]
            A dictionary of paths.
        """
        python_template_name = python_template_name or "default"
        return {
            'path_to_mod_file': self.get_abs_path(f'mod/{mechanism_name}.mod'),
            'path_to_python_file': self.get_abs_path(f'python/{mechanism_name}.py'),
            'path_to_python_template': self.get_abs_path(f'templates/{python_template_name}.py'),
        }


    def get_standard_channel_paths(self, mechanism_name: str,
                                   python_template_name: str = None,
                                   mod_template_name: str = None) -> Dict[str, str]:
        """
        Get all necessary paths for creating a standard channel.

        Parameters
        ----------
        mechanism_name : str
            The name of the mechanism.
        python_template_name : str, optional
            The name of the Python template file.
        mod_template_name : str, optional
            The name of the MOD template file.

        Returns
        -------
        Dict[str, str]
            A dictionary of paths.
        """
        python_template_name = python_template_name or "default"
        mod_template_name = mod_template_name or "standard_channel"
        return {
            # **self.get_channel_paths(mechanism_name, python_template_name),
            'path_to_mod_template': self.get_abs_path(f'templates/{mod_template_name}.mod'),
            'path_to_standard_mod_file': self.get_abs_path(f'mod/std{mechanism_name}.mod'),
        }


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

__version__ = "0.5.0"

from dendrotweaks.model import Model
from dendrotweaks.simulators import NeuronSimulator
from dendrotweaks.biophys.distributions import Distribution
from dendrotweaks.path_manager import PathManager
from dendrotweaks.stimuli import Synapse, Population, IClamp

from dendrotweaks.utils import download_example_data
from dendrotweaks.utils import apply_dark_theme


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

class Domain:
    """
    A class representing
    a morphological or functional domain in a neuron.

    Parameters
    ----------
    name : str
        The name of the domain.
    type_idx : int
        The type index of the domain.
    color : str
        The color of the domain.
    """

    def __init__(self, name, type_idx, color) -> None:
        self._name = name
        self._type_idx = type_idx
        self._color = color
        self._sections = []


    def __repr__(self):
        return f'<Domain({self.name}, {self.type_idx}, {self.color}, {len(self.sections)} sections)>'

    def __getitem__(self, idx):
        return self._sections[idx]

    def __len__(self):
        return len(self._sections)

    def __iter__(self):
        for sec in self._sections:
            yield sec

    def __contains__(self, sec):
        return sec in self._sections


    @property
    def sections(self):
        """
        A list of sections in the domain.
        """
        return self._sections


    @property
    def name(self):
        """
        The name of the domain.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        for sec in self._sections:
            for point in sec.points:
                point.domain_name = value


    @property
    def type_idx(self):
        """
        The type index of the domain.
        """
        return self._type_idx

    @type_idx.setter
    def type_idx(self, value):
        self._type_idx = value
        for sec in self._sections:
            for point in sec.points:
                point.type_idx = value

    
    @property
    def color(self):
        """
        The color of the domain.
        """
        return self._color

    
    @color.setter
    def color(self, value):
        self._color = value
        for sec in self._sections:
            for point in sec.points:
                point.domain_color = value


    def add_section(self, sec: "Section"):
        """
        Add a section to the domain.

        Changes the domain attribute of the section.

        Parameters
        ----------
        sec : Section
            The section to be added to the domain.
        """
        if sec in self._sections:
            warnings.warn(f'Section {sec} already in domain {self.name}.')
            return
        sec._domain = self
        sec.idx_within_domain = len(self._sections)
        for point in sec.points:
            point.domain_name = self.name
            point.type_idx = self.type_idx
            point.domain_color = self.color
        self._sections.append(sec)


    def remove_section(self, sec):
        """
        Remove a section from the domain.

        Sets the domain attribute of the section
        to None.

        Parameters
        ----------
        sec : Section
            The section to be removed from the domain.
        """
        if sec not in self.sections:
            warnings.warn(f'Section {sec} not in domain {self.name}.')
            return
        sec._domain = None
        sec.idx_within_domain = None
        for point in sec.points:
            point.domain_name = None
            point.type_idx = None
            point.domain_color = None
        if hasattr(sec, 'path_distance_within_domain'):
            # Remove cached property if it exists
            del sec.path_distance_within_domain
        self._sections.remove(sec)


    def is_empty(self):
        """
        Check if the domain is empty.

        Returns
        -------
        bool
            True if the domain is empty, False otherwise.
        """
        return not bool(self._sections)



# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

"""
Utility functions for dendrotweaks package.
"""

import time
import numpy as np
import os
import zipfile
import urllib.request
import matplotlib.pyplot as plt

DOMAINS_TO_GROUPS = {
    'soma': 'somatic',
    'axon': 'axonal',
    'dend': 'dendritic',
    'apic': 'apical',
}

DOMAINS_TO_NEURON = {
    'soma': 'soma',
    'perisomatic': 'dend_11',
    'axon': 'axon',
    'apic': 'apic',
    'dend': 'dend',
    'basal': 'dend_31',
    'trunk': 'dend_41',
    'tuft': 'dend_42',
    'oblique': 'dend_43',
    'custom': 'dend_5',
    'reduced': 'dend_8',
    'undefined': 'dend_0',
}


def timeit(func):
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(f"  Elapsed time: {round(end-start, 3)} seconds")
        return result
    return wrapper

def calculate_lambda_f(distances, diameters, Ra=35.4, Cm=1, frequency=100):
    """
    Calculate the frequency-dependent length constant (lambda_f) according to NEURON's implementation,
    using 3D point data for accurate representation of varying diameter frusta.
    
    Args:
        distances (list/array): Cumulative euclidean distances between 3D points along the section from 0 to section length
        diameters (list/array): Corresponding diameters at each position in micrometers
        Ra (float): Axial resistance in ohm*cm
        Cm (float): Specific membrane capacitance in µF/cm²
        frequency (float): Frequency in Hz
    
    Returns:
        float: Lambda_f in micrometers
    """
    if len(distances) < 2 or len(diameters) < 2:
        raise ValueError("At least 2 points are required for 3D calculation")
    
    if len(distances) != len(diameters):
        raise ValueError("distances and diameters must have the same length")
    
    # Initialize variables
    lam = 0
    section_L = distances[-1]
    
    # Calculate the contribution of each frustum
    for i in range(1, len(distances)):
        # Frustum length
        frustum_length = distances[i] - distances[i-1]
        # Average of diameters at endpoints
        d1 = diameters[i-1]
        d2 = diameters[i]
        
        # Add frustum contribution to lambda calculation
        lam += frustum_length / np.sqrt(d1 + d2)
    
    # Apply the frequency-dependent factor
    lam *= np.sqrt(2) * 1e-5 * np.sqrt(4 * np.pi * frequency * Ra * Cm)
    
    # Return section_L/lam (electrotonic length of the section)
    return section_L / lam

def dynamic_import(module_name, class_name):
    """
    Dynamically import a class from a module.

    Parameters
    ----------
    module_name : str
        Name of the module to import.
    class_name : str
        Name of the class to import.
    """

    from importlib import import_module

    import sys
    sys.path.append('app/src')
    print(f"Importing class {class_name} from module {module_name}.py")
    module = import_module(module_name)
    return getattr(module, class_name)

def list_folders(path_to_folder):
    folders = [f for f in os.listdir(path_to_folder)
            if os.path.isdir(os.path.join(path_to_folder, f))]
    sorted_folders = sorted(folders, key=lambda x: x.lower())
    return sorted_folders

def list_files(path_to_folder, extension):
    files = [f for f in os.listdir(path_to_folder)
            if f.endswith(extension)]
    return files

def write_file(content: str, path_to_file: str, verbose: bool = True) -> None:
    """
    Write content to a file.

    Parameters
    ----------
    content : str
        The content to write to the file.
    path_to_file : str
        The path to the file.
    verbose : bool, optional
        Whether to print a message after writing the file. The default is True.
    """
    if not os.path.exists(os.path.dirname(path_to_file)):
        os.makedirs(os.path.dirname(path_to_file))
    with open(path_to_file, 'w') as f:
        f.write(content)
    print(f"Saved content to {path_to_file}")

def read_file(path_to_file):
    with open(path_to_file, 'r') as f:
        content = f.read()
    return content

def download_example_data(path_to_destination, include_templates=True, include_modfiles=True):
    """
    Download and extract specific folders from the DendroTweaks GitHub repository:
    - examples/                <- from examples subfolder (always included)
    - examples/Templates/      <- from src/dendrotweaks/biophys/default_templates (optional)
    - examples/Default/        <- from src/dendrotweaks/biophys/default_mod (optional)

    Parameters
    ----------
    path_to_destination : str
        The path to the destination folder where the data will be downloaded and extracted.

    include_templates : bool, optional
        If True, also extract default_templates/ into examples/Templates/.

    include_modfiles : bool, optional
        If True, also extract default_mod/ into examples/Default/.
    """
    if not os.path.exists(path_to_destination):
        os.makedirs(path_to_destination)

    from dendrotweaks import __version__
    tag = f"v{__version__}"
    repo_url = f"https://github.com/Poirazi-Lab/DendroTweaks/archive/refs/tags/{tag}.zip"
    zip_path = os.path.join(path_to_destination, "dendrotweaks_repo.zip")

    print(f"Downloading data for v{__version__} from {repo_url}...")
    urllib.request.urlretrieve(repo_url, zip_path)

    print(f"Extracting relevant folders to {path_to_destination}...")
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        for member in zip_ref.namelist():
            target_path = None

            # === Always extract examples/ folder ===
            if member.startswith(f"DendroTweaks-{__version__}/examples/"):
                rel_path = os.path.relpath(member, f"DendroTweaks-{__version__}/examples")
                target_path = os.path.join(path_to_destination, rel_path)

            # === Optionally extract Templates/ folder ===
            elif include_templates and member.startswith(f"DendroTweaks-{__version__}/src/dendrotweaks/biophys/default_templates/"):
                rel_path = os.path.relpath(member, f"DendroTweaks-{__version__}/src/dendrotweaks/biophys/default_templates")
                target_path = os.path.join(path_to_destination, "Templates", rel_path)

            # === Optionally extract Default/ folder ===
            elif include_modfiles and member.startswith(f"DendroTweaks-{__version__}/src/dendrotweaks/biophys/default_mod/"):
                rel_path = os.path.relpath(member, f"DendroTweaks-{__version__}/src/dendrotweaks/biophys/default_mod")
                target_path = os.path.join(path_to_destination, "Default", rel_path)

            if target_path:
                if member.endswith('/'):
                    os.makedirs(target_path, exist_ok=True)
                else:
                    os.makedirs(os.path.dirname(target_path), exist_ok=True)
                    with zip_ref.open(member) as source, open(target_path, 'wb') as target:
                        target.write(source.read())

    os.remove(zip_path)
    print(f"Data downloaded and extracted successfully to {path_to_destination}/.")

def apply_dark_theme():
    """
    Apply a dark theme to matplotlib plots.
    """
    # dark theme
    plt.style.use('dark_background')

    # customize the style
    plt.rcParams.update({
        'figure.facecolor': '#131416',
        'axes.facecolor': '#131416',
        'axes.edgecolor': 'white',
        'axes.labelcolor': 'white',
        'xtick.color': 'white',
        'ytick.color': 'white',
        'text.color': 'white',
        'axes.prop_cycle': plt.cycler(color=plt.cm.tab10.colors),  # use standard matplotlib colors
    })

def mse(y_true, y_pred):
            return np.mean((np.array(y_true) - np.array(y_pred)) ** 2)

def max_error(y_true, y_pred):
    return np.max(np.abs(np.array(y_true) - np.array(y_pred)))

def poly_fit(x, y, max_degree=6, tolerance=1e-6):
    """
    Fit a polynomial to the data and return the coefficients and predicted values.
    """
    for degree in range(max_degree + 1):
        coeffs = np.polyfit(x, y, degree)
        y_pred = np.polyval(coeffs, x)
        if np.all(np.abs(np.array(y) - y_pred) < tolerance):
            break
    return coeffs, y_pred

def step_fit(x, y):
    """
    Fit a single step function with variable-width transition zone.
    Returns (high_val, low_val, start, end), and predicted y-values.
    """
    x = np.array(x)
    y = np.array(y)

    sort_idx = np.argsort(x)
    x = x[sort_idx]
    y = y[sort_idx]

    best_score = float('inf')
    best_params = None
    best_pred = None

    n = len(x)
    for i in range(n - 1):
        for j in range(i + 1, n):
            start = x[i]
            end = x[j]
            inside = (x > start) & (x < end)
            outside = ~inside

            if not np.any(inside) or not np.any(outside):
                continue

            high_val = np.nanmean(y[inside])
            low_val = np.nanmean(y[outside])

            if low_val >= high_val:
                continue

            pred = np.where(inside, high_val, low_val)
            score = max_error(y, pred)

            if score < best_score:
                best_score = score
                best_params = (start, end, low_val, high_val)
                best_pred = pred

    return best_params, best_pred

DEFAULT_FIT_MODELS = {
    'poly': {
        'fit': poly_fit,
        'score': max_error,
        'complexity': lambda coeffs: len(coeffs) - 1  # degree of polynomial
    },
    'step': {
        'fit': step_fit,
        'score': max_error,
        'complexity': lambda params: 4
    }
}


# SPDX-FileCopyrightText: 2025 Poirazi Lab <dendrotweaks@dendrites.gr>
# SPDX-License-Identifier: MPL-2.0

from dendrotweaks.biophys.io.loader import MODFileLoader
from dendrotweaks.biophys.io.converter import MODFileConverter

from dendrotweaks.biophys.io.reader import MODFileReader
from dendrotweaks.biophys.io.parser import MODFileParser
from dendrotweaks.biophys.io.code_generators import PythonCodeGenerator
from dendrotweaks.biophys.io.code_generators import NMODLCodeGenerator

from dendrotweaks.biophys.io.factories import create_channel
from dendrotweaks.biophys.io.factories import create_standard_channel
from dendrotweaks.biophys.io.factories import standardize_channel